; RSX for CP/M Plus (v3) to implement CP/Net.
; Built from DRI's NDOS.MAC and SNIOS.ASM for CPNetDevice emulation.
	maclib z80

;  NETWORK DISK OPERATING SYSTEM FOR CP/NET
;
;  1982.8.4. FOR BASE
;
;TITLE	NETWORK DISK OPERATING SYSTEM FOR CP/NET SLAVE
;
;
;  EQUATIONS OF DATA
;
LF	EQU	0AH		;LINE FEED
CR	EQU	0DH		;CARRIAGE RETURN
EOF	EQU	1AH		;CTRL-Z IS END OF FILE
;
FSEND	EQU	0		;SEND FORMAT CODE
FRECV	EQU	1		;RECEIVE FORMAT CODE
;
TOP	EQU	0		;TOP OF MEMORY
CDISK	EQU	4		;CURRENT DISK & USER CODE
BDOS	EQU	5		;BDOS ENTRY
SYSDMA	EQU	80H		;SYSTEM DMA ADDRESS
;
SCTLNG	EQU	128		;ONE SECTOR LENGTH
;
NOFUNC	EQU	8000H		;NO FUNCTION CODE IN CP/NET
;
;  EQUATIONS OF DOS FUNCTION
;
CCNDIN	EQU	3		;CONSOLE INPUT WITH DEVICE CODE
CCNDOT	EQU	4		;CONSOLE OUTPUT WITH DEVICE CODE
CBUFPR	EQU	9		;BUFFER PRINT
CCONST	EQU	11		;GET CONSOLE STATUS
CGETVR	EQU	12		;GET VERSION NUMBER
CRSDSK	EQU	13		;RESET DISK
COPEN	EQU	15		;OPEN FILE
CSRFST	EQU	17		;SEARCH FIRST
CSRNXT	EQU	18		;SEARCH NEXT DIRECTORY
CREAD	EQU	20		;READ FILE
CSTDMA	EQU	26		;SET DMA ADDRESS
CSTUSC	EQU	32		;SET USER CODE
CRSDSN	EQU	37		;RESET DISK BY DISK VERCTOR
;
CLOGIN	EQU	64		;LOGIN
CLOGOF	EQU	65		;LOGOFF
CNMAX	EQU	74		;MAX OF NDOS FUNCTION
;
;  SLAVE CONFIGRATION TABLE
;
; -1	NETWORK STATUS
;  0	SLAVE PROCESSOR ID
;  1-32	A - P DISK DEVICE CODE
; 33-34	CONSOLE DEVICE
; 35-36	LIST DEVICE
; 37	LIST BUFFER COUNTER
; 38-42	MESSAGE HEADER FOR LIST OUT
; 43	LISTER DEVICE NUMBER
; 44-171 LIST OUT DATA BUFFER
;
;  EACH DEVICE DATA USED 2 BYTES
;  IN 1-36
;  1B:BIT 7 H ON NETWORK
;     BIT 6 H SET UP IN DISK
;     BIT 0-3 DEVICE NUMBER IN MASTER
;  2B:MASTER ID
;
;
;  BIAS TO DATA IN CONFIGRATION TABLE
;
BSCONS	EQU	33		;BIAS TO CONSOLE DATA
BSLIST	EQU	35		;BIAS TO LISTER DATA

;	RSX Prefix
serial:	db	0,0,0,0,0,0
start:	jmp	COLDST
next:	jmp	0
prev:	dw	0
remove:	db	0	; 0ffh for remove
nonbank:
	db	0
rsxnam:	db	'CPNET3  '
loader:	db	0,0,0
; make it possible to patch I/O Port
ioport: db	018h

;
;  COMMENT
;
;
	DB	'COPYRIGHT (C) 1980, DIGITAL RESEARCH '
;
;  PARAMETER AREA
;
CONTAD:	DW	0		;CONFIGRATION TABLE ADDRESS
CURDSK:	DB	1		;CURRENT DISK
DMAADD:	DW	SYSDMA		;DMA ADDRESS
USRCOD:	DB	0		;USER CODE
VERSION:DW	0		;VERSION NUMBER
FLSRCH:	DB	0		;SEARCH DIRECTORY MODE  -1 ON NETWORK
PNSRCH:	DW	0		;FCB POINTER FOR SEARCH NEXT
;
;  MESSAGE DATA
;
MSGTOP:				;MESSAGE AREA TOP
	DB	0		;MESSAGE FORMAT CODE
MSGID:	DB	0		;DESTINATION DEVICE ID CODE
	DB	0		;SOURCE DEVICE ID CODE
MSGFUN:	DB	0		;FUNCTION CODE
MSGSIZ:	DB	0		;MESSAGE SIZE-1
MSGDAT:				;DATA
	DS	256		;MESSAGE DATA AREA
;
;  STACK AREA
;
	DB	0C7H,0C7H,0C7H,0C7H,0C7H,0C7H
	DB	0C7H,0C7H,0C7H,0C7H,0C7H,0C7H
	DB	0C7H,0C7H,0C7H,0C7H,0C7H,0C7H
	DB	0C7H,0C7H,0C7H,0C7H,0C7H,0C7H
	DB	0C7H,0C7H,0C7H,0C7H,0C7H,0C7H
	DB	0C7H,0C7H,0C7H,0C7H,0C7H,0C7H
	DB	0C7H,0C7H,0C7H,0C7H,0C7H,0C7H
	DB	0C7H,0C7H,0C7H,0C7H,0C7H,0C7H
	DB	0C7H,0C7H,0C7H,0C7H,0C7H,0C7H
	DB	0C7H,0C7H,0C7H,0C7H,0C7H,0C7H
STACK:
;
;  DOS WORKING
;
USTACK:	DW	0		;USER STACK SAVE BUFFER
FUNCOD:	DB	0		;FUNCTION CODE
PARAMT:	DW	0		;INPUT PARAMETER
MCRPNT:	DW	0		;MESSAGE CURRENT POINTER
LSTUNT:	DB	0		;LISTER UNIT IN MASTER
;
;  COMMENTS FOR COLD START
;
CSTUP:	DB	'NDOS initialization complete.',CR,LF,'$'
CLDERR:	db	'Failed to initialize CP/Net$'

COLDSE:
	push	b
	push	d
	lxi	d,CLDERR
	mvi	c,CBUFPR
	call	next
	lxi	h,next
	shld	start+1
	mvi	a,0ffh
	sta	remove
	pop	d
	pop	b
	jmp	next	; pass to BDOS and hope for the best...

WARMST:
	push	b
	call	ntwkbt
	pop	b
	jmp	next	; pass along WARM START

COLDST:
	push	b
	call	ntwkin
	pop	b
	ora	a
	jnz	COLDSE
	lxi	h,NDOSE
	shld	start+1
	xra	a
	sta	remove
	push	b
	push	d
	mvi	c,12
	call	next
	shld	VERSION
	lxi	d,CSTUP
	mvi	c,CBUFPR
	call	next
	pop	d
	pop	b
;
;  NDOS FUNCTION DECODING BODY
;
NDOSE:
	LXI	H,0
	MOV	A,C
	cpi	59
	jz	ldfunc
	CPI	CNMAX+1
	jnc	next		; do not presume to know BDOS limits
	DAD	SP
	SHLD	USTACK		;SAVE USER STACK
	LXI	SP,STACK	;SET NDOS STACK
	LXI	H,NDEND
	PUSH	H		;SET END ROUTINE POINTER
	MVI	A,CSRNXT
	SUB	C
	JNZ	NDOSES		;NOT SEARCH NEXT DIRECTORY
	LHLD	PNSRCH		;SEARCH NEXT DIRECTORY, SO SET UP FCB POINT
	SHLD	PARAMT
	LDA	FLSRCH
	ORA	A
	JNZ	NDOSEN		;ON NETWORK
	jmp	next		;LOCAL DISK - returns to NDEND
;
NDOSES:
	DCR	A
	JNZ	NDOSSS		;NOT SEARCH FIRST
	STA	FLSRCH		;SEARCH FIRST, SO FLAG CLEAR
NDOSSS:
	XCHG
	SHLD	PARAMT		;SAVE PARAMETER
NDOSEN:
	MOV	A,C
	STA	FUNCOD		;SAVE FUNCTION CODE
	STA	MSGFUN
	LXI	H,MSGSIZ
	MVI	M,0		;CLEAR SIZE
	INX	H
	SHLD	MCRPNT		;INITIALIZE MESSAGE POINTER
	MVI	B,0
	LXI	H,FNINDX
	DAD	B
	DAD	B
	DAD	B
	DAD	B		;GET INDEX POINT
	MOV	A,M
	MOV	E,A
	INX	H
	MOV	D,M		;GET ROUTINE POINTER
	ORA	D
	JZ	NDOSS2		;NOT ROUTINE TYPE
	XCHG			;ROUTINE TYPE
	PCHL			;TO EACH ROUTINE
NDOSS2:
	INX	H		;NOT ROUTINE TYPE
	INX	H
	MOV	A,M
	ANI	80H
	JNZ	TBDOSP		;NOT NDOS MODE, TO BODS
	XCHG			;BY BIT MAP PROCESS
	CALL	CHKDSK		;CHECK DISK & OVERRIDING GOTO BDOS, IF LOCAL
	XCHG
	DCX	H
	DCX	H
	MVI	B,2		;BYTE COUNT
	LXI	D,INBMAP	;BIT MAP INDEX TOP
NDOSBM:				;BYTE LOOP FOR SET UP
	INX	H
	MVI	C,8		;BIT COUNT
	MOV	A,M
NDOSBT:				;BIT LOOP
	RAR
	JNC	NDOSBS		;OFF BIT
	PUSH	H		;ON BIT
	PUSH	PSW
	LDAX	D
	MOV	L,A
	INX	D
	LDAX	D
	MOV	H,A		;GET ROUTINE POINT
	DCX	D
	POP	PSW
	XTHL			;PUSH ROUTINE ADDRESS
NDOSBS:
	INX	D
	INX	D		;TO NEXT INDEX POINT
	DCR	C
	JNZ	NDOSBT		;TO NEXT BIT
	DCR	B
	RZ			;OK ALL SET UP, TO FIRST PROCESS
	PUSH	H
	LXI	H,RCVPAR	;SET RECEIVE ROUTINE
	XTHL
	PUSH	H
	LXI	H,SNDHDR	;SET SEND ROUTINE
	XTHL
	JMP	NDOSBM		;TO PREPROCESS SET UP

; DEBUG messages
ldfmsg:	db	'Func 59 LOAD OVERLAY$'
ldfrmv:	db	' [REMOVE]$'
ldfnul:	db	' [NULL]$'
ldfend:	db	13,10,'$'

ldfunc:	; LOAD OVERLAY (and RSX remove)
	push	b
	push	d
; begin DEBUG code
	lxi	d,ldfmsg
	mvi	c,CBUFPR
	call	next
	lda	remove
	ora	a
	jz	ldf1
	lxi	d,ldfrmv
	mvi	c,CBUFPR
	call	next
ldf1:
	pop	d
	push	d
	mov	a,d
	ora	l
	jnz	ldf2
	lxi	d,ldfnul
	mvi	c,CBUFPR
	call	next
ldf2:
	lxi	d,ldfend
	mvi	c,CBUFPR
	call	next
; end DEBUG code
	pop	d
	pop	b
	jmp	next
;
;  SEND MESSAGE WITH HEADER SET UP
;
SNDHDR:
	LHLD	CONTAD
	XCHG
	LXI	H,MSGTOP
	MVI	M,FSEND		;FORMAT CODE TO SEND MODE
	LDAX	D		;GET SLAVE ID
	INX	H
	INX	H
	MOV	M,A		;SET SLAVE ID
	INX	H
	MOV	A,M
	CPI	CSRFST
	JNZ	SNDHDS		;NOT SEARCH FIRST
	MVI	A,-1
	STA	FLSRCH		;SET SEARCH FCB ON NETWORK MODE
	XCHG
	LHLD	PARAMT
	SHLD	PNSRCH		;SAVE FCB POINT
	XCHG
SNDHDS:
	INX	H
	INX	H
	XCHG
	LHLD	MCRPNT		;GET CURRENT POINT OF MESSAGE (= END POINT)
	XRA	A
	SUB	E
	MOV	C,A
	MVI	A,0
	SBB	D
	MOV	B,A
	DAD	B		;GET LENGTH OF MESSAGE
	MOV	A,L
	ORA	H
	JNZ	NDOSS3		;SOME DATA
	INX	H		;ZERO BECOMES ONE DATA
NDOSS3:
	DCX	H
	XCHG
	DCX	H
	MOV	M,E		;SET SIZE
	LXI	B,MSGTOP
	JMP	SDMSGE		;SEND MESSAGE
;
;  RECEIVE MESSAGE WITH INITIALIZATION OF MESSAGE POINTER
;
RCVPAR:
	LXI	B,MSGTOP
	CALL	RVMSGE		;RECEIVE
	LXI	H,MSGDAT
	SHLD	MCRPNT		;INITIALIZE POINTER
	RET
;
;  BIT MAP FUNCTION ROUTINES
;
;SUBTTL	BIT MAP FUNCTION ROUTINES
;
;  INDEX FOR BIT MAP FUNCTIONS OF DOS FUNCTION
;
INBMAP:
				;POST PROCESS ROUTINES
	DW	GTRPA		;0 GET RETURN PARAMETER
	DW	RSRRCN		;1 RESTORE RANDOM RECODE NUMBER
	DW	GTDRDT		;2 GET DIRECTORY DATA
	DW	GTOSCT		;3 GET ONE SECTOR DATA
	DW	RSEXCR		;4 RESTORE EXTENT & CURRENT RECORD
	DW	BMNOP		;5 NO OPERATION (NOT USED)
	DW	BMNOP		;6 NO OPERATION (NOT USED)
	DW	BMNOP		;7 NO OPERATION (NOT USED)
				;PRE PROCESS ROUTINES
	DW	BMNOP		;0 NO OPERATION (NOT USED)
	DW	STRRCN		;1 SET RANDOM RECODE NUMBER
	DW	STFLNM		;2 SET FILE NAME FOR RENAME
	DW	WTDTCP		;3 SET ONE SETOR DATA
	DW	STEXCR		;4 SET EXTENT & CURRENT RECODE
	DW	SETFCB		;5 SET FCB
	DW	STFCBP		;6 SET FCB POINT
	DW	TBDOSP		;7 TO BDOS (NOT USED)
;
;  NO OPERATION   POST 5,6,7  PRE 0	(NOT USED)
;
BMNOP:
	RET
;
;  POST OPERATIONS
;
;  POST 1
;  35:COMPUTE FILE SIZE
;  36:SET CURRENT RECODE NUMBER
;
;  RESTORE RANDOM RECORD NUMBER
;
RSRRCN:
	LHLD	PARAMT		;GET FCB TOP
	LXI	D,33		;BIAS TO RANDOM RECODE NUMBER
	DAD	D
	MVI	B,3
	JMP	MCPYFS		;COPY RANDOM RECODE NUMBER
;
;  POST 2
;  17:SEARCH FIRST
;  18:SEARCH NEXT
;
;  GET DIRECTORY DATA
;  OUTPUT
;   A:DIRECTORY CODE
;
GTDRDT:
	LHLD	MCRPNT
	XCHG
	LDAX	D
	INX	D
	CPI	-1
	RZ			;ERROR (NOT FOUND)
	LHLD	DMAADD
	LXI	B,32
	PUSH	PSW		;SAVE DIRECTORY CODE
	ANI	03H		;GET DIRECTORY BIAS CODE
	INR	A
NDOSS4:
	DCR	A
	JZ	GTDRDL		;GET BIAS
	DAD	B
	JMP	NDOSS4		;MAKE DATA TOP POINT
;
GTDRDL:				;COPY LOOP OF DIRECTORY DATA
	LDAX	D
	MOV	M,A
	INX	D
	INX	H
	DCR	C
	JNZ	GTDRDL
	POP	PSW		;RESTORE DIRECTORY CODE
	RET
;
;  POST 3
;  20:READ SEQUENTIAL
;  33:READ RANDOM
;
;  GET ONE SECTOR DATA
;
GTOSCT:
	LHLD	DMAADD
	MVI	B,SCTLNG
	JMP	MCPYFS		;COPY ONE SECTOR DATA TO DMA ADDRESS
;
;  POST 4
;  20:READ SEQUENTIAL
;  21:WRITE SEQUENTIAL
;  33:READ RANDOM
;  34:WRITE RANDOM
;  35:COMPUTE FILE SIZE
;  40:WRITE RANDOM WITH ZERO FILL
;
;  RESTORE EXTENT & CURRENT RECORD
;
RSEXCR:
	LHLD	MCRPNT
	XCHG
	LHLD	PARAMT		;FCB TOP
	LXI	B,12
	DAD	B
	LDAX	D
	MOV	M,A		;SET EXTENT
	INX	D
	LXI	B,20
	DAD	B
	LDAX	D
	MOV	M,A		;SET CURRENT RECODE
	INX	D
	XCHG
	SHLD	MCRPNT
	RET
;
;  POST 0
;  15:OPEN FILE
;  16:CLOSE FILE
;  19:DELETE FILE
;  20:READ SEQUENTIAL
;  21:WRITE SEQUENTIAL
;  22:MAKE FILE
;  23:RENAME
;  30:SET ATTRIBUTE
;  33:READ RANDOM
;  34:WRITE RANDOM
;  35:COMPUTE FILE SIZE
;  36:GET CURRENT RECODE NUMBER
;  40:WRITE RANDOM WITH ZERO FILL
;
;  GET RETURN PARAMETER
;  OUTPUT
;  HL:TWO BYTE PARAMETER
;   A:ONE BYTE PARAMETER  SAME WITH L
;
GTRPA:
	LHLD	MCRPNT
	MOV	E,M
	INX	H
	MOV	D,M		;GET PARAMETER
	INX	H
	SHLD	MCRPNT		;SAVE NEXT POINT
	XCHG
	MOV	A,L		;SET FOR ONE BYTE PARAMETER
	RET
;
;  PRE OPERATIONS
;
;  PRE 1
;  33:READ RANDOM
;  34:WRITE RANDOM
;  40:WRITE RANDOM WITH ZERO FILL
;
;  SET RANDOM RECORD NUMBER
;
STRRCN:
	LHLD	PARAMT		;FCB TOP
	LXI	D,33
	DAD	D		;RANDOM RECODE NUMBER POINT
	MVI	B,3
	JMP	MCPYTS
;
;  PRE 2
;  23:RENAME
;
;  SET FILE NAME FOR RENAME
;
STFLNM:
	LHLD	MCRPNT
	PUSH	H
	MVI	B,32
	CALL	MCPYTO		;SET TWO FILE NAMES
	CALL	CHKDSK		;GET DISK PARAMETER POINT
	DCX	H
	MOV	A,M
	ANI	0FH		;GET DISK NUMBER
	INR	A
	POP	H
	MOV	M,A		;SET DISK CODE
	LXI	D,16
	DAD	D
	MOV	M,A		;SET DISK CODE FOR SECOND NAME
	RET
;
;  PRE 3
;  21:WRITE SEQUENTIAL
;  34:WRITE RANDOM
;  40 WRITE RANDOM WITH ZERO FILL
;
;  SET ONE SECTOR DATA COPY TO MESSAGE BUFFER
;
WTDTCP:
	LHLD	DMAADD
	MVI	B,SCTLNG
	JMP	MCPYTS
;
	RET
;
;  PRE 4
;  20:READ SEQUENTIAL
;  21:WRITE SEQUENTIAL
;  36:GET CURRENT RECODE COUNT
;
;  SET EXTENT & CURRENT RECODE
;
STEXCR:
	LHLD	MCRPNT
	XCHG
	LHLD	PARAMT
	LXI	B,12
	DAD	B
	MOV	A,M		;SET EXTENT
	STAX	D
	INX	D
	LXI	B,20
	DAD	B
	MOV	A,M		;SET CURRENT RECORD
	STAX	D
	INX	D
	XCHG
	SHLD	MCRPNT
	RET
;
;  PRE 5
;  15:OPEN FILE
;  17:SEARCH FIRST
;  19:DELETE
;  22:MAKE FILE
;  30:SET ATTRIBUTE
;
;  SET FCB
;
SETFCB:
	LHLD	MCRPNT
	PUSH	H
	MVI	B,13
	CALL	MCPYTO		;COPY FCB
	CALL	CHKDSK
	DCX	H
	MOV	A,M
	ANI	0FH
	INR	A
	POP	H
	MOV	M,A		;SET DISK CODE
	RET
;
;  PRE 6
;  15:OPEN FILE
;  16:CLOSE FILE
;  20:READ SEQUENTIAL
;  21:WRITE SEQUENTIAL
;  22:MAKE FILE
;  33:READ RANDOM
;  34:WRITE RANDOM
;  35:COMPUTE FILE SIZE
;  36:GET CURRENT RECODE NUMBER
;  40:WRITE RANDOM WITH ZERO FILL
;
;  SET FCB POINT
;
STFCBP:
	LXI	H,PARAMT
	MVI	B,2
	JMP	MCPYTS
;
;  GENERAL SUBROUTINES
;
;SUBTTL	GENERAL SUBROUTINES
;
;  MESSAGE COPY FROM MESSAGE BUFFER
;  INPUT
;   B:COUNT
;
MCPYFR:				;NOT USED HERE
	LHLD	PARAMT		;GET DATA POINT
MCPYFS:
	XCHG
	LHLD	MCRPNT		;GET MESSAGE BUFFER POINT
NDOSS5:
	MOV	A,M		;COPY DATA
	STAX	D
	INX	D
	INX	H
	DCR	B
	JNZ	NDOSS5		;TO NEXT BYTE
	SHLD	MCRPNT		;SAVE NEXT POINT
	RET
;
;  MESSAGE COPY TO MESSAGE BUFFER
;  INPUT
;   B:COUNT
;
MCPYTO:
	LHLD	PARAMT		;GET PARAMETER POINT
MCPYTS:
	XCHG
	LHLD	MCRPNT		;GET MESSAGE POINT
NDOSS6:
	LDAX	D		;COPY DATA
	MOV	M,A
	INX	D
	INX	H
	DCR	B
	JNZ	NDOSS6
	SHLD	MCRPNT		;SAVE NEXT POINT
	RET
;
;  GET DISK PARAMETER POINT
;  INPUT
;   A:DISK NUMBER
;  OUTPUT
;  HL:PARAMETER POINT
;
GETDPT:
	LHLD	CONTAD
	DCX	H		;ADJUST A:DATA -2 POINT
	ADD	A		;GET BIAS
	ADD	L
	MOV	L,A
	MOV	A,H
	ACI	0
	MOV	H,A
	RET
;
;  CHECK & SET UP DISK & OVERRIDING GOTO BDOS, IF LOCAL
;  OUTPUT
;  DE:DE
;  HL:DISK PARAMETER POINT+1
;
CHKDSK:
	LHLD	PARAMT		;GET FCB POINT
	MOV	A,M		;GET ENTRY TYPE
	ORA	A
	JNZ	CHKDSS		;HAS ENTRY TYPE
	LDA	CURDSK
	INR	A		;MAKE ENTRY TYPE
CHKDSS:
	CALL	GETDPT		;GET DISK PARAMETER POINT
	MOV	A,M
	RAL
	JNC	TBDOSS		;LOCAL, SO OVERRIDING GOTO BDOS
	RAL
CHKDSL:
	INX	H
	MOV	A,M
	STA	MSGID		;SET MASTER ID
	RC			;ALREADY SET UP DISK
	DCX	H
	MVI	A,40H
	ORA	M		;ON SET UP BIT
	MOV	M,A
	PUSH	H
	PUSH	D
	MVI	A,CSTUSC
	STA	MSGFUN		;SET FUNCTION FOR SET UP USER CODE
	LDA	USRCOD
	STA	MSGDAT
	CALL	SNDHDR		;SET UP USER CODE
	CALL	RCVPAR
	LDA	FUNCOD
	LXI	H,MSGFUN
	MOV	M,A		;RESTOTRE FUNCTION CODE
	INX	H
	MVI	M,0		;CLEAR SIZE
	INX	H
	SHLD	MCRPNT		;ESTORE POINTER
	POP	D
	POP	H
	STC			;SET OK FLAG
	JMP	CHKDSL
;
;  TO BDOS WITH PARAMETER
;
TBDOSS:				;OVERRIDING TO BDOS
	POP	H
TBDOSP:
	LHLD	PARAMT		;GET PARAMETER
	XCHG
	LDA	FUNCOD		;GET FUNCTION CODE
	MOV	C,A
	jmp	next
;
;  SEND MESSAGE WITH ERROR PROCESS
;  INPUT
;  BC:MESSAGE ADDRESS
;
SDMSGE:
	CALL	SNDMSG		;SEND IN SNIOS
	INR	A
	RNZ			;NOT ERROR
	JMP	NERROR		;ERROR
;
;  RECEIVE MESSAGE WITH ERROR PROCESS
;  INPUT
;  BC:MESSAGE ADDRESS
;
RVMSGE:
	CALL	RCVMSG		;RECEIVE IN SNIOS
	INR	A
	RNZ			;NOT ERROR
;
;  ERROR END OF NDOS FUNCTION
;
NERROR:
	LXI	H,0FFFFH
	MOV	A,H
;
;  END OF DOS FUNCTION
;  INPUT
;  HL:RETURN PARAMETER
;   A:RETURN PARAMETER
;
NDEND:
	XCHG
	LHLD	USTACK		;GET USER STACK
	SPHL			;RESTORE USER STACK
	XCHG
	RET
;
;  BDOS & NDOS FUNCTION INDEX
;
;  2B:ROUTINE POINTER  0 BY BIT MAP
;  2B:BIT MAP	BIT 0-7   PRE PROCESS
;		BIT 8-14  POST PROCESS
;		BIT 15 H  NOT BY NDOS FUNCTION
;
FNINDX:
	DW	WARMST,0	; 0 SYSTEM RESET  WARM BOOT
	DW	0,NOFUNC	; 1 CONSOLE INPUT
	DW	0,NOFUNC	; 2 CONSOLE OUTPUT
	DW	0,NOFUNC	; 3 READER INPUT
	DW	0,NOFUNC	; 4 PUNCHER OUTPUT
	DW	0,NOFUNC	; 5 LISTER OUTPUT
	DW	0,NOFUNC	; 6 DIRECT CONSOLE I/O
	DW	0,NOFUNC	; 7 GET I/O BYTE
	DW	0,NOFUNC	; 8 SET I/O BYTE
	DW	0,NOFUNC	; 9 PRINT STRING
	DW	0,NOFUNC	;10 READ CONSOLE BUFFER
	DW	0,NOFUNC	;11 GET CONSOLE STATUS
	DW	GETVER,0	;12 GET VERSION NUMBER
	DW	0,NOFUNC	;13 RESET DISK
	DW	SELDSK,0	;14 SELECT DISK
;
	DW	0,6001H		;15 OPEN FILE
	DW	0,4001H		;16 CLOSE FILE
	DW	0,2004H		;17 SEARCH FIRST
	DW	0,0004H		;18 SEARCH NEXT
	DW	0,2001H		;19 DELETE FILE
	DW	0,5019H		;20 READ SEQUENTIAL
	DW	0,5811H		;21 WRITE SEQUENTIAL
	DW	0,6001H		;22 MAKE FILE
	DW	0,0401H		;23 RENAME
	DW	0,NOFUNC	;24 GET LOGIN VECTOR
	DW	GETCDK,0	;25 GET CURRENT DISK
	DW	SETDMA,0	;26 SET DMA ADDRESS
	DW	0,NOFUNC	;27 GET ALLOCATION VECTOR
	DW	SETROD,0	;28 SET TO R/O DISK
	DW	0,NOFUNC	;29 GET R/O VECTOR
	DW	0,2001H		;30 SET ATTRIBUTE
	; TBD: should this be implemented?
	DW	0,NOFUNC	;31 GET DISK PARAMETER ADDRESS
	DW	SETUSR,0	;32 SET USER CODE
	DW	0,4219H		;33 READ RANDOM
	DW	0,4A11H		;34 WRITE RANDOM
	DW	0,4013H		;35 COMPUTE FILE SIZE
	DW	0,5003H		;36 GET CURRENT RECODE NUMBER
	DW	DSKACC,0	;37 RESET DISK
	DW	DSKACC,0	;38 ACCESS DISK
	DW	DSKACC,0	;39 FREE DISK
	DW	0,4A11H		;40 WRITE RANDOM WITH ZERO FILL
	; For CP/M 3 (and MP/M) these must be passed along to BDOS
	DW	0,NOFUNC	;41 ----SKIP---- MP/M
	DW	0,NOFUNC	;42 ----SKIP---- MP/M
	DW	0,NOFUNC	;43 ----SKIP---- MP/M
	; TBD: can multi-sector-count be implemented for network drives?
	DW	0,NOFUNC	;44 ----SKIP---- Multi-sector count
	DW	0,NOFUNC	;45 ----SKIP---- BDOS Err mode
	; TBD: can disk free space be implemented for network drives?
	DW	0,NOFUNC	;46 ----SKIP---- Disk Free Space
	DW	0,NOFUNC	;47 ----SKIP---- Chain to Prog
	DW	0,NOFUNC	;48 ----SKIP---- Flush buffers
	DW	0,NOFUNC	;49 ----SKIP---- Get/set SCB
	DW	0,NOFUNC	;50 ----SKIP---- Direct BIOS
	DW	0,NOFUNC	;51 ----SKIP----
	DW	0,NOFUNC	;52 ----SKIP----
	DW	0,NOFUNC	;53 ----SKIP----
	DW	0,NOFUNC	;54 ----SKIP----
	DW	0,NOFUNC	;55 ----SKIP----
	DW	0,NOFUNC	;56 ----SKIP----
	DW	0,NOFUNC	;57 ----SKIP----
	DW	0,NOFUNC	;58 ----SKIP----
	DW	0,NOFUNC	;59 ----SKIP---- Load overlay (RSX unload)
	DW	0,NOFUNC	;60 ----SKIP---- Call RSX
	DW	0,NOFUNC	;61 ----SKIP---- 
	DW	0,NOFUNC	;62 ----SKIP----
	DW	0,NOFUNC	;63 ----SKIP----
	DW	LOGIN,0		;64 LOGIN
	DW	LOGOFF,0	;65 LOGOFF
	DW	SENDMS,0	;66 SEND MESSAGE
	DW	RECVMS,0	;67 RECEIVE MESSAGE
	DW	NTWKST,0	;68 NETWORK STATUS
	DW	CNFTBL,0	;69 GET CONFIGRATION TABLE
	DW	FUNOVR,0	;70 ----OVER----
	DW	FUNOVR,0	;71 ----OVER----
	DW	FUNOVR,0	;72 ----OVER----
	DW	FUNOVR,0	;73 ----OVER----
	DW	FUNOVR,0	;74 ----OVER----
;
;  NDOS FUNCTION ROUTINES
;
;SUBTTL	NDOS FUNCTION ROUTINES
;
;  GET VERSION NUMBER
;  OUTPUT
;  HL:VERSION NUMBER
; For reference:
;                H      L
; CP/M 1.4   00000000B 00H (FUNC 12 = "Lift Head": succes return code)
; CP/M 2.2   00000000B 22H
; CP/M 3.1   00000000B 31H * come overloading, e.g. Concurrent CP/M is xx31H
; MP/M II    00000001B 30H
; CP/Net     xxxxxx1xB xxH (Client/Slave) x = underlying local OS version
; ConcCP/M   xxxxx1x0B xxH
; i8080      0000xxxxB xxH (TBD: exact bit encoding)
; i8086      0001xxxxB xxH * some docs suggest 0000xxxxB, but prob wrong
; m68K       0010xxxxB xxH
; z8000      0011xxxxB xxH
;
GETVER:
	LHLD	VERSION		;GET ORIGINAL VERSION NUMBER
	MVI	A,02H
	ORA	H		;TO CP/NET VERSION
	MOV	H,A
	MOV	A,L
	RET
;
;  SELECT DISK
;
SELDSK:
	LDA	PARAMT
	STA	CURDSK		;SAVE NEW DISK
;
;  SET R/O DISK
;
SETROD:
	LDA	PARAMT
	INR	A
	CALL	CHKDSS		;CHECK DISK & TO BDOS IF LOCAL
	DCX	H
	MOV	A,M		;GET MASTER DISK
	ANI	00FH
	STA	MSGDAT		;SET DISK
	CALL	SNDHDR		;SEND
	CALL	RCVPAR		;RECEIVE
	JMP	GTRPA		;GET RETURN PARAMETER
;
;  GET CURRENT DISK
;  OUTPUT
;   A:CURRENT DISK
;
GETCDK:
	LDA	CURDSK
	RET
;
;  SET DMA ADDRESS
;
SETDMA:
	LHLD	PARAMT
	SHLD	DMAADD
	JMP	TBDOSP		;TO BDOS, ALSO SET UP IN BDOS
;
;  GET OR SET USER CODE
;
SETUSR:
	LDA	PARAMT
	CPI	-1
	MOV	B,A
	LDA	USRCOD
	RZ			;INTEROGATE MODE
	CMP	B
	RZ			;SAME CODE
	MOV	A,B
	ANI	0FH
	STA	USRCOD		;SAVE NEW CODE
	LHLD	CONTAD
	MVI	C,16		;DISK COUNT
NDOSS7:
	INX	H
	MOV	A,M
	ANI	0BFH		;RESET USER CODE SET UP MODE OF EACH DISK
	MOV	M,A
	INX	H
	DCR	C
	JNZ	NDOSS7
	LHLD	PARAMT
	XCHG
	LDA	VERSION
	CPI	20H		;CHECK VERSION
	RC			;TOO LOWER VERSION
	MVI	C,CSTUSC
	JMP	next		;SET USER CODE
;
;  SET UP DISK ACCESS MODE
;  RESET DISK, ACCESS DISK & FREE DISK
;
DSKACC:
	LHLD	CONTAD
	LXI	D,31
	DAD	D		;GET DISK DATA BOTTOM
	XCHG
	LXI	H,0
	MVI	C,16
DSKACM:				;MAKE NETWORK DISK VECTOR
	DAD	H
	LDAX	D
	DCX	D
	DCX	D
	RAL
	JNC	NDOSS8		;LOACL DISK
	INX	H		;NETWORK DISK  SET VECTOR BIT
NDOSS8:
	DCR	C
	JNZ	DSKACM		;TO NEXT CHANEL
	XCHG
	LHLD	PARAMT
	MOV	A,E
	ANA	L
	MOV	L,A
	MOV	A,D
	ANA	H
	MOV	H,A		;GET DISK VECTOR ON NETWORK
	ORA	L
	JNZ	DSKANT		;SOME DISK ON NETWORK
DSKACL:				;MAKE PROCESS FOR LOCAL DISK
	MOV	A,E
	CMA
	MOV	E,A
	MOV	A,D
	CMA
	LHLD	PARAMT
	ANA	H
	MOV	H,A
	MOV	A,L
	ANA	E
	MOV	L,A		;OFF BIT OF NETWORK DISK
	SHLD	PARAMT
	LDA	VERSION
	CPI	20H
	JNC	TBDOSP		;TO BDOS
	LDA	FUNCOD		;TOO LOWER VERSION
	CPI	CRSDSN		;CHECK DISK RESET
	MVI	C,CRSDSK	;RESET ALL DISK CODE
	JZ	next		;TO RESET ALL DISK
	RET			;NO PROCESS
;
DSKANT:				;RESET WITH NETWORK DISK
	PUSH	H
	CALL	DSKACL		;RESET LOCAL DISK
	MVI	C,16
	LHLD	CONTAD
	LXI	D,31
	DAD	D		;GET BOTTOM OF DISK PARAMETER
	POP	D
	XCHG
DSKANL:				;NETWORK DISK RESET LOOP
	DAD	H
	JNC	DSKANS		;NOT RESET DISK
	PUSH	B
	PUSH	D
	PUSH	H
	XCHG
	MOV	A,M
	ANI	0FH		;GET DISK IN MASTER
	MOV	C,A
	INR	C
	INX	H
	MOV	A,M		;GET MASTER ID
	LXI	H,1
NDOSD2:
	DCR	C
	JZ	NDOSD1		;GET BIT POINT OF DISK
	DAD	H
	JMP	NDOSD2		;UP BIT POINT
NDOSD1:
	XCHG
	LXI	H,MSGID		;SET MASTER ID
	MOV	M,A
	INX	H
	INX	H
	LDA	FUNCOD
	MOV	M,A		;SET FUNCTION
	LXI	H,MSGDAT
	MOV	M,E
	INX	H
	MOV	M,D		;SET DISK VECTOR
	INX	H
	SHLD	MCRPNT
	CALL	SNDHDR		;SEND
	CALL	RCVPAR		;RECEIVE
	CALL	GTRPA		;GET RETURN PARAMETER
	POP	H
	POP	D
	POP	B
	CPI	-1
	RZ			;ERROR
DSKANS:
	DCX	D
	DCX	D
	DCR	C
	JNZ	DSKANL		;TO NEXT DISK
	XRA	A		;SUCCESS END
	RET
;
;  ERROR OF FUNCTION CODE
;
FUNERR:
	LXI	H,0FFFFH
	MOV	A,H
	RET
;
;  LOGIN
;
LOGIN:
	LHLD	PARAMT
	PUSH	H
	INX	H		;GET PASSWORD POINT
	MVI	B,8		;PASSWORD LENGTH
	CALL	MCPYTS		;SET PASSWORD
	POP	H
	MOV	A,M		;GET MASTER ID
	MVI	B,CLOGIN	;FUNCTION CODE
	JMP	LOGCOM
;
;  LOGOFF
;
LOGOFF:
	MVI	B,CLOGOF	;FUNCTION CODE
	LDA	PARAMT		;MASTER ID
LOGCOM:				;LOGIN LOGOFF COMMON
	LXI	H,MSGID
	MOV	M,A		;SET MASTER ID
	INX	H
	INX	H
	MOV	M,B		;GET FUNCTION
	CALL	SNDHDR
	CALL	RCVPAR
	JMP	GTRPA
;
;  SEND MESSAGE FOR NDOS FUNCTION
;
SENDMS:
	LHLD	PARAMT
	MOV	B,H
	MOV	C,L		;GET MESSAGE BUFFER POINT
	JMP	SNDMSG
;
;  RECEIVE MESSAGE FOR NDOS FUNCTION
;
RECVMS:
	LHLD	PARAMT
	MOV	B,H
	MOV	C,L		;GET MESSAGE BUFFER POINT
	JMP	RCVMSG
;
;  FUNCTION CODE ERROR  TOO LARGE CODE
;
FUNOVR:
	LXI	H,0FFFFH
	MOV	A,H
	RET
;
;
;;;;;;; END of DRI code ;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; SNIOS code to use the emulator CPNetDevice, a pseudo device for CP/Net messages.

defaultmaster	equ	00h

;	Slave Configuration Table
configtbl:
nwst:	ds	1		; network status byte
	ds	1		; slave processor ID number
	ds	2		; A:  Disk device
	ds	2		; B:   "
	ds	2		; C:   "
	ds	2		; D:   "
	ds	2		; E:   "
	ds	2		; F:   "
	ds	2		; G:   "
	ds	2		; H:   "
	ds	2		; I:   "
	ds	2		; J:   "
	ds	2		; K:   "
	ds	2		; L:   "
	ds	2		; M:   "
	ds	2		; N:   "
	ds	2		; O:   "
	ds	2		; P:   "

	ds	2		; console device

	ds	2		; list device:
	ds	1		;	buffer index
	db	0		;	FMT
	db	0		;	DID
	db	0ffh		;	SID (CP/NOS must still initialize)
	db	5		;	FNC
	ds	1		;	SIZ
	ds	1		;	MSG(0)	List number
	ds	128		;	MSG(1) ... MSG(128)

;	Network Status Byte Equates
;
active		equ	0001$0000b	; slave logged in on network
rcverr		equ	0000$0010b	; error in received message
senderr 	equ	0000$0001b	; unable to send message

;	Utility Procedures
;
;	Network Initialization
ntwkin:
	lxix	configtbl
	mvi	a,active
	stx	a,0 ; network status byte
	lda	ioport
	mov	c,a
	inr	c
	xra	a
	outp	a
	dcr	c
	inp	a
	stx	a,1 ; our slave (client) ID
	xra	a
	ret

;	Network Status
ntwkst:
	lda	nwst
	mov	b,a
	ani	not (rcverr+senderr)
	sta	nwst
	mov	a,b
	ret

;	Return Configuration Table Address
;	Still need this for BDOS func 69
cnftbl:
	lxi	h,configtbl
	ret

;	Send Message on Network
sndmsg:			; BC = message addr
	mov	h,b
	mov	l,c	; HL = message address
	push	h
	popix
	lda	ioport
	mov	c,a
	mvi	b,5	; length of header
	outir
	ldx	b,4	; msg siz field (-1)
	inr	b  	; might be 0, but that means 256
	outir
	inr	c	; status port
	inp	a	;
	ani	02h	; cmd overrun
	rz
	mvi	a,0ffh
	ret

;	Receive Message from Network
rcvmsg:			; BC = message addr
	mov	h,b
	mov	l,c	; HL = message address
	push	h
	popix
	lda	ioport
	mov	c,a
	inr	c	; status port
rcvwait:
	inp	a
	ani	01h	; data ready
	jz	rcvwait
	dcr	c
	mvi	b,5	; header length
	inir
	ldx	b,4	; msg siz
	inr	b
	inir
	inr	c
	inp	a
	ani	04h	; rsp overrun
	rz
	mvi	a,0ffh
	ret

ntwkbt:	; WARM START
	lda	ioport
	mov	c,a
	outp	a
	ret

	end
