	maclib z80

BDOS    equ     0005h

BUFLEN	equ	128

; standard BDOS calls
WBOOT   equ     0
CONIN   equ     1
CONOUT  equ     2
PRINT   equ     9
GTLINE  equ     10
GETVER  equ     12
OPEN    equ     15
CLOSE   equ     16
SERFST  equ     17
SERNXT  equ     18
DELETE  equ     19
SREAD   equ     20
SWRITE  equ     21
MAKE    equ     22
RENAME  equ     23
SETDMA  equ     26
SGUSER  equ     32
RREAD   equ     33
RWRITE  equ     34
FSIZE   equ     35
ERRMOD  equ     45
GTTIME	equ	105

; CP/NET (NDOS) calls
NCFGTB  equ     69
NSVCFG  equ     71

; MP/M (XDOS) calls
XPRIO   equ     145
XSDPA   equ     154
; Offsets in SDP
SDPCPN	equ	9	; CP/Net master config table
SDPTMP	equ	196

; Offsets in Server Config Table (as used by CP/Net Clients)
SCTTMP	equ	0	; Temp file drive
SCTSTS	equ	1	; Network (device) status (MP/M SDP entry points here)
SCTNID	equ	2	; Node ID
SCTMAX	equ	3	; Max allowed logins
SCTCUR	equ	4	; Num logged in
SCTVEC	equ	5	; (word) Login vector
SCTRID	equ	7	; 16x1 Table of remote node IDs logged in
SCTLEN	equ	23	; Length of table

; Offsets in Client Config Table
NCTSTS	equ	0	; Network (device) status
NCTNID	equ	1	; Node ID
NCTMAP	equ	2	; 16x2 Map of disk drives
NCTCON	equ	34	; Map for CON:
NCTLST	equ	36	; Map for LST:
NCTIDX	equ	38	; Index into LST: buffer
NCTMSG	equ	39	; CP/NET message buffer for 129-byte payload (LST: buffer)
NCTLEN	equ	173	; Length of table

; Offsets in standard FCB
FCBDRV	equ	0	; Disk drive, 1-16, or 0 for current default
FCBNAM	equ	1	; 8x NAME begins
FCBTYP	equ	9	; 3x TYPE begins
FCBEXT	equ	12	; EXT byte
FCBS1	equ	13	; S1
FCBS2	equ	14	; S2
FCBRC	equ	15	; RC - Record Count of extent
FCBD0	equ	16	; 16x D0-D15, internal use only
FCBREN	equ	FCBD0	; Second name entry for RENAME
FCBCR	equ	32	; CR - current record (within extent)
FCBRR0	equ	33	; 3x RR - random record number
FCBLEN	equ	36	; Length of full FCB

CR	equ	13
LF	equ	10
EOF	equ	26	; End Of Message marker, plus text EOF

; MAIL constants
MSGMXR	equ	13	; max records for MAIL message, excl. subject/header
MSGTYP	equ	0	; Record type - 0AAH, 000H/01AH.
MSGSIZ	equ	1	; Record data len - includes "from" on rec 0
MSGHDR	equ	2	; overhead in records 1-N. Data begins here.
MSGMAX	equ	BUFLEN-MSGHDR	; max length of data in record
; Record 0 - Subject and Envelope information
MSGFRM	equ	2	; From: ID
MSGHD0	equ	3	; record 0 - subject text begins here
MSGTIM	equ	4	; time/date at end of record
MSGMXS	equ	BUFLEN-MSGHD0-MSGTIM	; max length of Subject text
; Record types
MSGSOM	equ	0aah	; Start Of Message marker (first/subject record)
MSGEOM	equ	EOF	; End Of Message marker (last record)
MSGDAT	equ	0	; Middle Record(s) marker.
MSGDEL	equ	0ffh	; Deleted Record(s) marker (until next packfile).

	org	0100h
	jmp	start

	db	'MAIL Version 1.2'
	db	'COPYRIGHT 1982, DIGITAL RESEARCH'
	db	'080282'

start:
	lxi	sp,stack
	mvi	e,0ffh
	mvi	c,ERRMOD
	call	BDOS
	mvi	e,0ffh
	mvi	c,SGUSER
	call	BDOS
	sta	curusr
	mvi	c,GETVER
	call	BDOS
	shld	version
	mov	a,h
	ani	01h
	jnz	mpminit
	mov	a,h
	ani	02h
	jnz	cpninit
; CP/M is an error... for now...
	lxi	d,cpmerr
exit0:
	mvi	c,PRINT
	call	BDOS
exit1:
	lda	curusr
	mov	e,a
	mvi	c,SGUSER
	call	BDOS
	mvi	c,WBOOT
	jmp	BDOS	; no return
;----------------------------------

mpminit:
	mvi	c,XSDPA
	call	BDOS
	mov	e,l
	mov	d,h
	lxi	b,SDPCPN
	dad	b
	mov	a,m
	inx	h
	mov	h,m
	mov	l,a
	push	h	; MP/M Net Srv Cfg Table
	lxi	h,SDPTMP
	dad	d
	mov	a,m
	sta	srvcfg+SCTTMP
	; BUG: by copying table, we use stale data
	; when searching logged-in nodes. Problem
	; increases the longer the user is in MAIL.
	lxi	h,srvcfg+SCTSTS
	pop	d
	lxi	b,22
	ldir
	lda	srvcfg+SCTNID
	sta	myid
	mvi	e,100
	mvi	c,XPRIO
	call	BDOS
	jmp	main

cpninit:
	mvi	c,NCFGTB
	call	BDOS
	shld	netcfg
	inx	h
	mov	a,m
	sta	myid
main:
	lxi	h,msgbuf
	mvi	m,MSGEOM
	inx	h
	mvi	m,0
toploop:
	mvi	a,1
	call	getopt
	ora	a
	jz	bcast
	dcr	a
	jz	send
	dcr	a
	jz	recv
	dcr	a
	jz	exit1
; error?
	jmp	toploop

;;;;;;;; BROADCAST A MESSAGE ;;;;;;;;;;
bcast:
	call	getsubject
	mvi	a,2
	call	getopt
	ora	a
	jz	bcast1
	call	gettext
	jmp	bcast2
bcast1:
	call	getinputfile
bcast2:
	lda	version+1
	ani	02h
	jnz	bccpnet

bcmpm:
	lda	srvcfg+SCTTMP
	sta	mailfcb+FCBDRV
	mvi	b,255	; exclude node - none
	call	bcast5
	jmp	toploop

bccpnet:
bccpn0:
	call	findtmp	; sets up srvcfg
	ora	a
	jz	toploop
	cpi	0ffh
	jz	bccpn0
	sta	mailfcb+FCBDRV
	lda	myid
	mov	b,a	; exclude node - ourself
	call	bcast5
	jmp	bccpn0

bcast5:
	lxi	h,srvcfg+SCTRID
	xchg
	lhld	srvcfg+SCTVEC
	mvi	c,16
bcmpm0:
	mov	a,l
	ani	1
	jz	bcmpm2
	ldax	d	; target node ID
	cmp	b
	jz	bcmpm2
	push	h
	push	d
	push	b
	call	getmailfile	; A still target node ID
	ora	a
	jnz	op1error	; mailfcb
	call	addmsg
	pop	b
	pop	d
	pop	h
bcmpm2:
	rarr	h
	rarr	l
	inx	d
	dcr	c
	jnz	bcmpm0
	ret

;;;;;;;; SEND A MESSAGE ;;;;;;;;;;
send:
	call	getnodeid
	sta	tonode
	call	getsubject
	mvi	a,2
	call	getopt
	ora	a
	jz	send1
	call	gettext
	jmp	send2
send1:
	call	getinputfile
send2:
	lda	version+1
	ani	02h
	jnz	sncpnet
snmpm:
	lda	srvcfg+SCTTMP
	sta	mailfcb+FCBDRV
	lda	tonode
	call	getmailfile
	ora	a
	jnz	op1error	; mailfcb
	call	addmsg
	jmp	toploop

sncpnet:

	; Locate first server or logged-in node matching dest.
	; This code may not be thorough enough.
sncpn0:
	call	findtmp
	ora	a
	jz	sncpn4
	cpi	0ffh
	jz	sncpn0
	sta	mailfcb+FCBDRV
	lda	tonode
	mov	b,a
	lda	srvcfg+SCTNID
	cmp	b
	jnz	sncpn1
	; send to server
	mov	a,b
	call	getmailfile
	cpi	0ffh
	jz	op1error
	call	addmsg
	jmp	sncpn6	; cleanup
	; Search server for logged-in node
sncpn1:
	lxi	h,srvcfg+SCTRID
	xchg
	lhld	srvcfg+SCTVEC
	mvi	c,16
	lda	tonode
	mov	b,a
sncpn2:
	mov	a,l
	ani	1
	jz	sncpn3
	ldax	d	; target node ID
	cmp	b
	jnz	sncpn3
	mov	a,b
	call	getmailfile	; A still target node ID
	ora	a
	jnz	op1error	; mailfcb
	call	addmsg
	jmp	sncpn6	; cleanup
sncpn3:
	rarr	h
	rarr	l
	inx	d
	dcr	c
	jnz	sncpn2
	; need to look for more servers...
	jmp	sncpn0

	; node not logged in (and not a server),
	; find any server with temp drive and deliver there.
sncpn4:
	call	findtmp
	ora	a
	jz	sncpn5
	cpi	0ffh
	jz	sncpn4
	sta	mailfcb+FCBDRV
	lda	tonode
	call	getmailfile
	ora	a
	jnz	op1error
	call	addmsg
sncpn6:	; search until end so findtmp does cleanup
	call	findtmp
	ora	a
	jnz	sncpn6
	jmp	toploop
sncpn5:
	lxi	d,notmps
	jmp	exit0

;;;;;;;; RECEIVE MAIL ;;;;;;;;;;
recv:
	lda	version+1
	ani	02h
	jnz	rcvcpn
	lda	srvcfg+SCTTMP
	sta	mailfcb+FCBDRV	; localtemp set in FCB
	jmp	rcvmpm
rcvcpn:
recv0:
	lda	version+1
	ani	02h
	jz	toploop	; MP/M stops here
	call	findtmp
	ora	a
	jz	recv8	; no cleanup needed, already done in case of zero....
	cpi	0ffh
	jz	recv0
	sta	mailfcb+FCBDRV	; localtemp set in FCB
rcvmpm:
	; localtemp, myid are set elsewhere
	call	viewmail
	ora	a
	jz	recv0
	dcr	a
	jz	toploop
recv8:
	lxi	d,nomore
	mvi	c,PRINT
	call	BDOS
	jmp	toploop

; B = myid, C = localtmp
; localtmp already set in mailfcb+FCBDRV...
; myid is set.
viewmail:
	lxi	h,0
	shld	mrec0
	shld	mrec1
vmail0:
	lda	myid
	call	getmailfile
	ora	a
	jnz	op1error	; mailfcb
	call	rdmail
	ora	a
	jnz	rstop2
	mvi	c,CONIN
	call	BDOS
	mvi	a,3
	call	getopt
	ora	a
	jz	rstop	; 1: Stop Receiving Mail
	dcr	a
	jz	ransw	; 2: Answer Message
	dcr	a
	jz	rdelt	; 3: Delete Message
	dcr	a
	jz	ransdl	; 4: Answer and Delete
	dcr	a
	jnz	vmail0	; 6: Next Message
	lhld	mrec0	; 5: Re-examine current message
	shld	mrec1
	jmp	vmail0

rstop2:	; EOF on mail file
	call	packfile
	mvi	a,2
	ret

rstop:	; 1: Stop Receiving Mail
	call	packfile
	mvi	a,1
	ret

ransw:	; 2: Answer Message
	call	replymsg
	jmp	vmail0

rdelt:	; 3: Delete Message
	call	delmsg
	jmp	vmail0

ransdl:	; 4: Answer and Delete
	call	replymsg
	call	delmsg
	jmp	vmail0

; end viewmail

; Show current menu and get option selection...
; A=currrent menu level (1, 2, 3)
getopt:
	dcr	a
	push	psw
	lxi	d,crlf2
	mvi	c,PRINT
	call	BDOS
	lxi	d,crlf2
	mvi	c,PRINT
	call	BDOS
	pop	psw
	push	psw
	add	a
	mov	e,a
	mvi	d,0
	lxi	h,mnutab
	dad	d
	mov	e,m
	inx	h
	mov	d,m
	mvi	c,PRINT
	call	BDOS
	lxi	d,prompt
	mvi	c,PRINT
	call	BDOS
	mvi	a,2
	sta	cmdline+0
	lxi	d,cmdline
	mvi	c,GTLINE
	call	BDOS
	pop	psw
	mov	e,a
	mvi	d,0
	lxi	h,numopt
	dad	d
	lda	cmdline+1
	cpi	1
	jnz	getop1
	lda	cmdline+2
	sui	'0'
	ora	a
	jz	getop1
	cmp	m
	jnc	getop1
	dcr	a
	ret
getop1:
	mov	a,m
	dcr	a
	ret

; Print filename of FCB in DE.
prfile:
	ldax	d	; drive, 1..16
	dcr	a
	adi	'A'
	push	d
	mov	e,a
	mvi	c,CONOUT
	call	BDOS
	mvi	e,':'
	mvi	c,CONOUT
	call	BDOS
	pop	d
	mvi	c,8
	call	prfil0
	push	d
	mvi	e,'.'
	mvi	c,CONOUT
	call	BDOS
	pop	d
	mvi	c,3
	call	prfil0
	lxi	d,crlf
	mvi	c,PRINT
	jmp	BDOS

prfil0:
	ldax	d
	push	d
	push	b
	mov	e,a
	mvi	c,CONOUT
	call	BDOS
	pop	b
	pop	d
	inx	d
	dcr	c
	jnz	prfil0
	ret

op2error:
	lxi	d,msgfcb
	jmp	opnerror
op1error:
	lxi	d,mailfcb
opnerror:
	lxi	h,operrm
errexit:
	push	d
	xchg
	mvi	c,PRINT
	call	BDOS
	pop	d
	call	prfile
	jmp	exit1

rd2error:
	lxi	d,msgfcb
	jmp	rderror
rd1error:
	lxi	d,mailfcb
rderror:
	lxi	h,rderrm
	jmp	errexit

wr2error:
	lxi	d,msgfcb
	jmp	wrerror
wr1error:
	lxi	d,mailfcb
wrerror:
	lxi	h,wrerrm
	jmp	errexit
	

; Return:
;  0:  No (more) servers,
;  0ffh:   Error, or temp drive not mapped
;  1-16:   Temp drive is local drive A-P
; Local vars:
tmpdrv:	db	0
tmploc:	db	0
curdrv:	db	0
tmpff:	db	0

findtmp:
	mvi	a,0ffh
	sta	tmploc
	sta	tmpdrv
	mov	b,a	; cursrv
	mvi	c,020h	; tmpflg
	xra	a
	sta	curdrv
	lhld	netcfg
	inx	h	;
	inx	h	; +NCTMAP
findt0:
	mov	a,m
	ani	0c0h
	ora	c
	sta	tmpff
	cpi	0a0h
	jnz	findt1
	mvi	c,0
	inx	h
	mov	b,m	; cursrv
	dcx	h
	push	h
	push	b
	mov	e,b
	mvi	c,NSVCFG
	call	BDOS
	mov	a,l
	cpi	0ffh
	jnz	findt5
	mov	a,h
	cpi	00ch
	jnz	findt5
	mvi	a,0ffh
	pop	b
	pop	h
	ret
findt5:	; HL=valid server config tbl
	lxi	d,srvcfg
	lxi	b,SCTLEN
	ldir
	lda	srvcfg+SCTTMP	; server tmp drive
	sta	tmpdrv
	pop	b
	pop	h
findt1:
	lda	tmpff
	ani	0c0h
	cpi	080h
	jnz	findt2
	inx	h
	mov	a,m
	dcx	h
	cmp	b
	jnz	findt2
	mvi	a,040h
	ora	m
	mov	m,a
	ani	0fh
	mov	e,a
	lda	tmpdrv
	cmp	e
	jnz	findt2
	lda	curdrv
	inr	a
	sta	tmploc
findt2:
	inx	h
	inx	h
	lda	curdrv
	inr	a
	sta	curdrv
	cpi	16
	jc	findt0
	mov	a,b
	cpi	0ffh
	jnz	findt3
	; cleanup - end of servers
findinit:
	xra	a
	sta	tmploc
	sta	curdrv
	lhld	netcfg
	inx	h	;
	inx	h	; +NCTMAP
	mvi	c,16
findt4:	; cleanup - reset temp flags
	mvi	a,08fh
	ana	m
	mov	m,a
	inx	h
	inx	h
	dcr	c
	jnz	findt4
	xra	a
	ret	; no (more) servers
findt3:
	lda	tmploc
	ret

; A = target node ID, drive set in mailfcb
getmailfile:
	; don't overwrite FCBDRV, already setup...
	lxi	h,mailtex+FCBNAM
	lxi	d,mailfcb+FCBNAM
	lxi	b,FCBLEN-FCBNAM
	ldir
	push	psw
	rrc
	rrc
	rrc
	rrc
	call	tohex
	sta	mailfcb+FCBNAM
	pop	psw
	call	tohex
	sta	mailfcb+FCBNAM+1
	lxi	d,dskbuf
	mvi	c,SETDMA
	call	BDOS
	lxi	d,mailfcb
	mvi	c,SERFST
	call	BDOS
	mvi	c,OPEN
	cpi	0ffh
	jnz	getmf1
	mvi	c,MAKE
getmf1:
	lxi	d,mailfcb
	call	BDOS
	cpi	0ffh
	rnz
	lxi	h,48
getmf2:
	push	h
	lxi	d,mailfcb
	mvi	c,OPEN
	call	BDOS
	pop	h
	cpi	0ffh
	rnz
	dcx	h
	mov	a,l
	ani	03h
	cpi	1
	jnz	getmf3
	push	h
	lxi	d,waitm
	mvi	c,PRINT
	call	BDOS
	pop	h
getmf3:
	mov	a,l
	ora	h
	jnz	getmf2
	mvi	a,0ffh
	ret

; Used by getinputfile(internally) and gettext/nxtbyte(shared)
bytcnt:	db	0
; Purely local to getinputfile
dptr:	dw	0

getinputfile:
	lda	curusr
	mov	e,a
	mvi	c,SGUSER
	call	BDOS
	lxi	d,l1dd8h
	mvi	c,PRINT
	call	BDOS
	mvi	a,15
	sta	cmdline+0
	lxi	d,cmdline
	mvi	c,GTLINE
	call	BDOS
	lxi	h,cmdline+2
	lda	cmdline+1
	mov	e,a
	mvi	d,0
	inx	h
	dad	d
	mvi	m,0	; NUL terminate
	lxi	h,blankfile
	lxi	d,msgfcb
	lxi	b,FCBLEN
	ldir
	mvi	c,8
	lxi	d,msgfcb+FCBNAM
	lxi	h,cmdline+2
	lda	cmdline+3
	cpi	':'
	jnz	getif1
	mov	a,m
	inx	h
	inx	h
	sui	'A'
	inr	a
	sta	msgfcb+FCBDRV
getif1:
	mov	a,m
	ora	a
	jz	getif2
	inx	h
	cpi	'.'
	jz	getif2
	stax	d
	inx	d
	dcr	c
	jnz	getif1
getif2:
	mvi	c,3
getif3:
	mov	a,m
	ora	a
	jz	getif4
	inx	h
	stax	d
	inx	d
	dcr	c
	jnz	getif3
getif4:
	lxi	d,msgfcb
	mvi	c,OPEN
	call	BDOS
	cpi	0ffh
	jz	op2error
	lxi	d,dskbuf
	mvi	c,SETDMA
	call	BDOS

	mvi	c,MSGMAX	; C=offset
	mvi	b,MSGMXR	; max reccnt
	lxi	d,msgbuf+BUFLEN	; DE=recptr
	lxi	h,2
	dad	d	; HL=ptr
	xra	a
	sta	bytcnt
getif5:
	push	d
	push	b
	push	h
	lxi	d,msgfcb
	mvi	c,SREAD
	call	BDOS
	ora	a
	jnz	getif6	; must pop!
	pop	h
	pop	b
	push	b
	push	h
	lxi	d,dskbuf
	xchg
	mvi	b,0
	ldir		; HL=(dptr += offset)
	shld	dptr
	lxi	h,bytcnt
	inr	m
	inr	m	; bytcnt += 2
	pop	h
	pop	b
	pop	d
	mvi	a,MSGDAT
	stax	d	; rec type 0 - more follows...
	dcr	b
	jz	getif7	; overflow
	lxi	h,BUFLEN
	dad	d
	xchg		; recptr += BUFLEN;
	lxi	h,2
	dad	d	; ptr = recptr + 2;
	dcr	c
	dcr	c	; offset -= 2;
	push	d
	push	b
	xchg		; DE=ptr
	lhld	dptr
	lda	bytcnt
	mov	c,a
	mvi	b,0
	ldir		; DE=(ptr+=bytcnt)
	xchg		; HL=ptr
	pop	b
	pop	d
	jmp	getif5
getif7:
	mvi	a,MSGEOM
	stax	d
	inx	d
	mvi	a,MSGMAX
	stax	d
getif9:
	lxi	d,msgfcb
	mvi	c,CLOSE
	call	BDOS
	mvi	e,0
	mvi	c,SGUSER
	jmp	BDOS
getif6:	; EOF or error...
	cpi	1
	jnz	rd2error	; no need to pop...
	; look for ASCII EOF...
	pop	h	; ptr
	pop	b	; offset
	pop	d	; recptr
	xchg
	dcr	h	; HL -= 256
	push	h
	lxi	d,2
	dad	d
	lxi	b,MSGMAX
	mvi	a,EOF
	ccir
	pop	h
	jnz	getif8
	mov	m,a
	inx	h
	mvi	a,MSGMAX
	sub	c
	mov	m,a
	jmp	getif9
getif8:
	lxi	d,BUFLEN
	dad	d
	push	h
	inx	h
	mov	c,m	; length
	mvi	b,0
	inx	h
	mvi	a,EOF
	ccir
	pop	h
	mov	m,a
	inx	h
	mvi	a,MSGMAX
	sub	c	; C could be 0
	mov	m,a
	jmp	getif9

; Used by gettext/nxtbyte to share context
reccnt:	db	0
recptr:	dw	0
msgptr:	dw	0

; A = bytcnt, returns new bytcnt
nxtbyte:
	lda	bytcnt
	inr	a
	sta	bytcnt	; not needed if end of rec...
	lhld	msgptr
	inx	h
	shld	msgptr	; not needed if end of rec...
	cpi	MSGMAX
	rc	; within same record, keep going.
	lhld	recptr
	mvi	m,MSGDAT
	lda	reccnt	; include header/subject record
	inr	a
	cpi	MSGMXR+1
	jnc	nxtby1	; overflow
	sta	reccnt	; point to next record
	lxi	d,BUFLEN
	dad	d
	shld	recptr
	inx	h
	inx	h
	shld	msgptr
	xra	a
	sta	bytcnt
	ret
nxtby1:
	mvi	m,MSGEOM
	inx	h
	mvi	m,MSGMAX
	mvi	a,0ffh
	ret

gettext:
	lxi	d,entmsg
	mvi	c,PRINT
	call	BDOS
	lxi	d,crlf
	mvi	c,PRINT
	call	BDOS
	mvi	a,78
	sta	cmdline+0
	mvi	a,1
	sta	reccnt
	lxi	h,msgbuf+BUFLEN
	shld	recptr
	inx	h
	inx	h
	shld	msgptr
	xra	a
	sta	bytcnt
gettx0:
	lxi	d,cmdline
	mvi	c,GTLINE
	call	BDOS
	lxi	d,crlf
	mvi	c,PRINT
	call	BDOS
	lxi	h,cmdline+1
	mov	e,m
	mvi	d,0
	inx	h
	dad	d
	mvi	m,0	; NUL terminate
	lxi	h,cmdline+2
gettx1:
	mov	a,m
	ora	a
	jz	gettx2
	cpi	EOF
	jz	gettx9
	inx	h
	push	h
	lhld	msgptr
	mov	m,a
	call	nxtbyte
	cpi	0ffh
	pop	h
	jnz	gettx1
	ret	; overflow
gettx2:
	push	h
	lhld	msgptr
	mvi	m,CR
	call	nxtbyte
	cpi	0ffh
	pop	h
	rz
	push	h
	lhld	msgptr
	mvi	m,LF
	call	nxtbyte
	cpi	0ffh
	pop	h
	rz
	jmp	gettx0
gettx9:
	lhld	recptr
	mvi	m,MSGEOM
	inx	h
	lda	bytcnt
	mov	m,a
	ret

getsubject:
	mvi	a,80
	sta	cmdline+0
	lxi	d,l1debh
	mvi	c,PRINT
	call	BDOS
	lxi	d,cmdline
	mvi	c,GTLINE
	call	BDOS
	lxi	h,msgbuf
	mvi	m,MSGSOM
	inx	h
	lda	cmdline+1
	mov	c,a
	inr	a
	mov	m,a
	inx	h
	lda	myid
	mov	m,a
	inx	h
	xchg
	lxi	h,cmdline+2
	mvi	b,0
	ldir
	lxi	h,msgbuf+BUFLEN-MSGTIM
	call	stdate
	ret

addmsg:
	lxi	d,mailfcb
	mvi	c,FSIZE
	call	BDOS
	cpi	0ffh
	jz	wr1error
	lxi	h,msgbuf
	mvi	c,MSGMXR+1
addmg0:
	push	h
	push	b
	xchg
	mvi	c,SETDMA
	call	BDOS
	lxi	d,mailfcb
	mvi	c,RWRITE
	call	BDOS
	ora	a
	jnz	wr1error
	lhld	mailfcb+FCBRR0
	inx	h
	shld	mailfcb+FCBRR0
	pop	b
	pop	h
	mov	a,m
	cpi	MSGEOM
	jz	addmg1
	lxi	d,BUFLEN
	dad	d
	dcr	c
	jnz	addmg0
addmg1:
	lxi	d,mailfcb
	mvi	c,CLOSE
	call	BDOS
	cpi	0ffh
	jz	wr1error
	ret

; assume < 100
decout:
	mvi	b,'0'
decot0:
	sui	10
	jc	decot1
	inr	b
	jmp	decot0
decot1:
	adi	10
	adi	'0'
	push	psw
	mov	a,b
	call	prout
	pop	psw
	jmp	prout

hexout:
	push	psw
	rrc
	rrc
	rrc
	rrc
	call	hexdig
	pop	psw
hexdig:
	call	tohex
prout:
	mov	e,a
	mvi	c,CONOUT
	jmp	BDOS

tohex:
	ani	0fh
	adi	90h
	daa
	aci	40h
	daa
	ret

; Keeps number in HL - caller must preserve/init
; Returns CY for invalid
hexnum:
	sui	'0'
	rc
	cpi	9+1
	jnc	hexnm1
hexnm2:
	dad	h
	dad	h
	dad	h
	dad	h
	ora	l
	mov	l,a
	ret
hexnm1:
	sui	'A'-'9'
	rc
	cpi	5+1
	cmc
	rc
	adi	10
	jmp	hexnm2


; prompt for node ID and parse input into hex number
getnodeid:
	mvi	a,3
	sta	cmdline+0
getni0:
	lxi	d,l1dffh
	mvi	c,PRINT
	call	BDOS
	lxi	d,cmdline
	mvi	c,GTLINE
	call	BDOS
	lda	cmdline+1
	ora	a
	jz	getni2
	cpi	3
	jnc	getni2
	lxi	h,0
	lxi	d,cmdline+2
	lda	cmdline+1
	mov	c,a
getni1:
	ldax	d
	inx	d
	call	hexnum
	jc	getni2
	dcr	c
	jnz	getni1
	mov	a,l
	ret
getni2:
	lxi	d,l1e12h
	mvi	c,PRINT
	call	BDOS
	jmp	getni0

; HL=string, C=length
strout:
	mov	a,c
	ora	a
	rz
strou0:
	mov	e,m
	push	h
	push	b
	mvi	c,CONOUT
	call	BDOS
	pop	b
	pop	h
	inx	h
	dcr	c
	jnz	strou0
	ret

; Load one message into msgbuf and display it
; Message to load is from record number mrec1
rdmail:
	lxi	d,mailfcb
	mvi	c,FSIZE
	call	BDOS
	cpi	0ffh
	jz	rd1error
	lhld	mrec1
	xchg
	lhld	mailfcb+FCBRR0
	ora	a
	dsbc	d
	jz	rdmai9
	xchg
	shld	mailfcb+FCBRR0
	lxi	h,msgbuf	; recptr
	mvi	c,MSGMXR+1
rdmai0:
	push	h
	push	b
	xchg
	mvi	c,SETDMA
	call	BDOS
	lxi	d,mailfcb
	mvi	c,RREAD
	call	BDOS
	ora	a
	jnz	rd1error
	lhld	mailfcb+FCBRR0
	inx	h
	shld	mailfcb+FCBRR0
	pop	b
	pop	h
	mov	a,m
	cpi	MSGEOM
	jz	rdmai1
	lxi	d,BUFLEN
	dad	d
	dcr	c
	jnz	rdmai0
	jmp	rd1error

rdmai1:
	lhld	mrec1
	shld	mrec0
	lhld	mailfcb+FCBRR0
	shld	mrec1
	lxi	d,mailfcb
	mvi	c,CLOSE
	call	BDOS
	cpi	0ffh
	jz	rd1error
rdmai2:
	; TODO handle status bit(s)...
	lda	msgbuf
	cpi	MSGSOM
	jnz	rd1error
	lxi	d,crlf2
	mvi	c,PRINT
	call	BDOS
	lxi	d,l1e2ah
	mvi	c,PRINT
	call	BDOS
	lda	msgbuf+MSGFRM
	call	hexout
	; TODO handle date/time
	lxi	d,l1e3eh
	mvi	c,PRINT
	call	BDOS
	lxi	h,msgbuf+BUFLEN-4
	call	prdate
	lxi	d,l1e3dh
	mvi	c,PRINT
	call	BDOS
	lxi	h,msgbuf+MSGHD0
	lda	msgbuf+MSGSIZ
	dcr	a	; don't count From ID
	mov	c,a
	call	strout
	lxi	d,crlf2
	mvi	c,PRINT
	call	BDOS
	lxi	h,msgbuf+BUFLEN	; recptr
	mvi	c,MSGMXR
rdmai4:
	push	h
	push	b
	mov	a,m	; MSGTYP
	inx	h
	cpi	MSGEOM
	jnz	rdmai5
	mov	c,m	; MSGSIZ
	jmp	rdmai6
rdmai5:
	mvi	c,MSGMAX
rdmai6:
	inx	h
	call	strout
	pop	b
	pop	h
	mov	a,m	; MSGTYP
	cpi	MSGEOM
	mvi	a,0
	rz
	lxi	d,BUFLEN
	dad	d	; HL (recptr) += BUFLEN
	dcr	c
	jnz	rdmai4
	jmp	rd1error
rdmai9:
	lxi	d,mailfcb
	mvi	c,CLOSE
	call	BDOS
	cpi	0ffh
	jz	rd1error
	mvi	a,0ffh
	ret

delmsg:
	lda	myid
	call	getmailfile
	cpi	0ffh
	jz	op1error
	lxi	h,msgbuf
	mvi	m,MSGDEL
	lxi	d,msgbuf+1
	lxi	b,BUFLEN-1
	ldir	; memset entire record - no shadow data
	lxi	d,msgbuf
	mvi	c,SETDMA
	call	BDOS
	lhld	mrec0
	shld	mailfcb+FCBRR0
delmg0:
	lxi	d,mailfcb
	mvi	c,RWRITE
	call	BDOS
	ora	a
	jnz	wr1error
	lhld	mailfcb+FCBRR0
	inx	h
	shld	mailfcb+FCBRR0
	xchg
	lhld	mrec1
	xchg
	ora	a
	dsbc	d
	jc	delmg0
	lxi	d,mailfcb
	mvi	c,CLOSE
	call	BDOS
	cpi	0ffh
	jz	wr1error
	ret

replymsg:
	lxi	h,msgbuf+MSGMXS-restrl
	lxi	d,msgbuf+MSGMXS
	lxi	b,MSGMXS-restrl
	lddr
	lxi	h,restr
	lxi	d,msgbuf+MSGHD0
	lxi	b,restrl
	ldir
	lxi	h,msgbuf+BUFLEN-MSGTIM
	call	stdate
	lxi	h,msgbuf+MSGSIZ
	mvi	a,restrl
	add	m
	cpi	MSGMXS+1	; length includes "from" byte
	jc	rpymg1
	mvi	a,MSGMXS+1
rpymg1:
	mov	m,a
	call	gettext
	lda	msgbuf+MSGFRM	; "From" node ID
	mov	b,a
	lda	myid
	sta	msgbuf+MSGFRM
	mov	a,b
	call	getmailfile
	cpi	0ffh
	jz	op1error
	call	addmsg
	ret

packfile:
	xra	a
	sta	mailfcb+FCBCR
	sta	mailfcb+FCBEXT
	sta	mailfcb+FCBS2
	lxi	h,mailfcb
	lxi	d,msgfcb
	lxi	b,FCBLEN
	ldir
	lxi	h,tmpsuf
	lxi	d,mailfcb+FCBTYP
	lxi	b,3
	ldir
	lxi	d,dskbuf
	mvi	c,SETDMA
	call	BDOS
	lxi	d,msgfcb
	mvi	c,OPEN
	call	BDOS
	cpi	0ffh
	jz	op2error
	lxi	d,mailfcb
	mvi	c,MAKE
	call	BDOS
	cpi	0ffh
	jz	op1error
pkfil0:
	lxi	d,msgfcb
	mvi	c,SREAD
	call	BDOS
	ora	a
	jnz	pkfil1
	lda	dskbuf+MSGTYP
	cpi	MSGDEL
	jz	pkfil0
	lxi	d,mailfcb
	mvi	c,SWRITE
	call	BDOS
	ora	a
	jnz	wr1error
	jmp	pkfil0
pkfil1:
	cpi	1
	jnz	rd2error
	lxi	d,mailfcb
	mvi	c,CLOSE
	call	BDOS
	cpi	0ffh
	jz	wr1error
	lxi	d,msgfcb
	mvi	c,DELETE
	call	BDOS
	cpi	0ffh
	jz	rd2error
	lxi	h,msgfcb
	lxi	d,mailfcb+FCBREN
	lxi	b,16
	ldir
	lxi	d,mailfcb
	mvi	c,RENAME
	call	BDOS
	cpi	0ffh
	jz	wr1error
	ret

; HL = CP/M Date-time field, w/o seconds
stdate:
	lda	version
	cpi	030h
	jc	stdat1
	xchg
	mvi	c,GTTIME
	jmp	BDOS
; Could try to contact network server(s) for time...
stdat1:
	xra	a
	mov	m,a
	inx	h
	mov	m,a
	inx	h
	mov	m,a
	inx	h
	mov	m,a
	ret

; HL = CP/M Date-time field, w/o seconds
; Print date and time to console.
prdate:
	mov	e,m
	inx	h
	mov	d,m
	inx	h
	push	h
; compute year
	mvi	c,78	; base year, epoch, binary
	mvi	b,078h	; year, BCD
	; special-case date=0...
	mov	a,e
	ora	d
	jnz	prdat0
	inx	d
prdat0:
	lxi	h,365
	mov	a,c
	ani	03h	; Not strictly true, but works until year 2100...
	jnz	prdat1
	inx	h
prdat1:	push	h
	ora	a
	dsbc	d
	pop	h
	jnc	prdat2	; done computing year...
	xchg
	ora	a
	dsbc	d
	xchg
	inr	c
	mov	a,b
	adi	1
	daa
	mov	b,a
	jmp	prdat0
prdat2:	; DE = days within year 'C'
	push	b	; save (2-digit) year, B = BCD, C = binary (until 2155)
	lxi	h,month0+24
	mov	a,c
	ani	03h
	jnz	prdat3
	lxi	h,month1+24
prdat3:	; compute month, DE = days in year,HL = mon-yr-days table adj for leap
	mvi	b,12
prdat4:
	dcx	h
	dcx	h
	dcr	b
	jm	prdat5	; should never happen...
	push	h
	push	d
	mov	a,m
	inx	h
	mov	h,m
	mov	l,a
		; DE = days in year, HL = ytd[month]
	ora	a
	dsbc	d
	mov	a,l	; potential remainder (neg)
	pop	d
	pop	h
	jnc	prdat4
prdat5:	; B = month, 0-11; A = -dom
	neg
	push	psw
	inr	b
	mov	a,b
	call	decout
	mvi	e,'/'
	mvi	c,CONOUT
	call	BDOS
	pop	psw
	call	decout
	mvi	e,'/'
	mvi	c,CONOUT
	call	BDOS
	pop	b
	mov	a,b	; already BCD
	call	hexout
	mvi	e,' '
	mvi	c,CONOUT
	call	BDOS
	pop	h	; -> BCD hours
	mov	a,m
	inx	h
	push	h
	call	hexout
	mvi	e,':'
	mvi	c,CONOUT
	call	BDOS
	pop	h
	mov	a,m
	jmp	hexout

;		J   F   M   A   M   J   J   A   S   O   N   D
month0:	dw	 0, 31, 59, 90,120,151,181,212,243,273,304,334
month1:	dw	 0, 31, 60, 91,121,152,182,213,244,274,305,335

; These need to be globals...
version	dw	0
srvcfg:	ds	SCTLEN
netcfg:	dw	0
curusr:	db	0
myid:	db	0
tonode:	db	0

; used only as template - never written or used as fcb
mailtex: db	0,'xxMAIL  TEX',0,0,0,0
	db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	db	0,0,0,0
blankfile: db	0,'           ',0,0,0,0
	db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	db	0,0,0,0

; Keeps track of current/next message for Receive Mail
mrec0:	dw	0
mrec1:	dw	0

waitm:	db	CR,LF,'Waiting for File$'
nomore:	db	CR,LF,CR,LF,'No More Messages$'
l1dd8h:	db	CR,LF,'Input Filename: $'
entmsg:	db	CR,LF,'Enter message:',CR,LF,'$'
l1debh:	db	CR,LF,'Enter Subject:'
prompt:	db	CR,LF,'>$'
l1dffh:	db	CR,LF,'Destination ID: $'
l1e12h:	db	CR,LF,'ILLEGAL DESTINATION',CR,LF,'$'
l1e2ah:	db	CR,LF,'Mail From Node #$'
l1e3dh:	db	CR,LF,'Subject: $'
l1e3eh:	db	CR,LF,'Date: $'
crlf2:	db	CR,LF
crlf:	db	CR,LF,'$'

restr:	db	'RE: '	; pre-pended to subject each time a message answered
restrl	equ	$-restr

notmps:	db	CR,LF,'SERVER MAIL DRIVE NOT NETWORKED',CR,LF,'$'
operrm:	db	CR,LF,'ERROR OPENING FILE $'
rderrm:	db	CR,LF,'ERROR READING FILE $'
wrerrm:	db	CR,LF,'ERROR WRITING FILE $'
tmpsuf:	db	'$$$'


mnumain: db	CR,LF,'Mail Program Main Menu',CR,LF    ;1ec2 0d .
	db	CR,LF,'Select one of the following:'
	db	CR,LF,'        1 - Broadcast'
	db	CR,LF,'        2 - Send Mail'
	db	CR,LF,'        3 - Receive Mail'
	db	CR,LF,'        4 - Exit From Program'
	db	'$'

mnusend: db	CR,LF,'Select source of message:'   ;1f66 0d .
	db	CR,LF,'        1 - File'
	db	CR,LF,'        2 - Console Input'
	db	'$'

mnurecv: db	CR,LF,'Select response:'
	db	CR,LF,'        1 - Stop Receiving Mail'
	db	CR,LF,'        2 - Answer Message'
	db	CR,LF,'        3 - Delete Message From Mail File'
	db	CR,LF,'        4 - Answer Message, Then Delete'
	db	CR,LF,'        5 - Re-Examine Last Message'
	db	CR,LF,'        6 - Get Next Message'
	db	'$'

cpmerr:	db	CR,LF,'Must Run Under MP/M or CP/NET.',CR,LF,'$'

mnutab:	dw	mnumain
	dw	mnusend
	dw	mnurecv
numopt:	db	4,2,6   ; number of options in menu

?bss	equ	$

	ds	64
stack:	ds	0

msgfcb:	ds	FCBLEN
mailfcb: ds	FCBLEN
cmdline: ds	80+2
dskbuf:	ds	BUFLEN
msgbuf:	ds	(MSGMXR+1)*BUFLEN

	org	?bss	; uninitialized data in .bss
	end
