	maclib z80
NIOS:
	public	NIOS
;	Jump vector for SNIOS entry points
	jmp	ntwrkinit	; network initialization
	jmp	ntwrksts	; network status
	jmp	cnfgtbladr	; return config table addr
	jmp	sendmsg 	; send message on network
	jmp	receivemsg	; receive message from network
	jmp	ntwrkerror	; network error
	jmp	ntwrkwboot	; network warm boot

defaultmaster	equ	00h

	DSEG

;	Slave Configuration Table
configtbl:

Network$status:
	ds	1		; network status byte
	ds	1		; slave processor ID number
	ds	2		; A:  Disk device
	ds	2		; B:   "
	ds	2		; C:   "
	ds	2		; D:   "
	ds	2		; E:   "
	ds	2		; F:   "
	ds	2		; G:   "
	ds	2		; H:   "
	ds	2		; I:   "
	ds	2		; J:   "
	ds	2		; K:   "
	ds	2		; L:   "
	ds	2		; M:   "
	ds	2		; N:   "
	ds	2		; O:   "
	ds	2		; P:   "

	ds	2		; console device

	ds	2		; list device:
	ds	1		;	buffer index
	db	0		;	FMT
	db	0		;	DID
	db	0ffh		;	SID (CP/NOS must still initialize)
	db	5		;	FNC
	ds	1		;	SIZ
	ds	1		;	MSG(0)	List number
	ds	128		;	MSG(1) ... MSG(128)

msg$adr:
	ds	2		; message address

;	Network Status Byte Equates
;
active		equ	0001$0000b	; slave logged in on network
rcverr		equ	0000$0010b	; error in received message
senderr 	equ	0000$0001b	; unable to send message

print	equ	9		; print string function

ioport: db	018h

	CSEG
;	Utility Procedures
;
	page
;	Network Initialization
ntwrkinit:
	lxix	configtbl
	mvi	a,active
	stx	a,0 ; network status byte
	lda	ioport
	mov	c,a
	inr	c
	xra	a
	outp	a
	dcr	c
	inp	a
	stx	a,1 ; our slave (client) ID
	xra	a
	ret

;	Network Status
ntwrksts:
	lda	network$status
	mov	b,a
	ani	not (rcverr+senderr)
	sta	network$status
	mov	a,b
	ret



;	Return Configuration Table Address
cnfgtbladr:
	lxi	h,configtbl
	ret


;	Send Message on Network
sendmsg:			; BC = message addr
	mov	h,b
	mov	l,c		; HL = message address
	push	h
	popix
	lda	ioport
	mov	c,a
	mvi	b,5 ; length of header
	outir
	ldx	b,4 ; msg siz field (-1)
	inr	b   ; might be 0, but that means 256
	outir
	inr	c ; status port
	inp	a ;
	ani	02h ; cmd overrun
	rz
	mvi	a,0ffh
	ret

;	Receive Message from Network
receivemsg:			; BC = message addr
	mov	h,b
	mov	l,c		; HL = message address
	push	h
	popix
	lda	ioport
	mov	c,a
	inr	c  ; status port
rcvwait:
	inp	a
	ani	01h ; data ready
	jz	rcvwait
	dcr	c
	mvi	b,5 ; header length
	inir
	ldx	b,4 ; msg siz
	inr	b
	inir
	inr	c
	inp	a
	ani	04h ; rsp overrun
	rz
	mvi	a,0ffh
	ret

ntwrkerror:
				;  perform any required device 
	ret			;     re-initialization

;
ntwrkwboot:
 
;	This procedure is called each time the CCP is
;	reloaded from disk.
	ret

	end
