import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.io.*;

public class H89SASIFormatter {
	// Taken from GenericSASIDrive. Keep in sync!
	public static final int XEBEC_ST506 = 1;	//  5.3MB (XEBEC ctrlr + Seagate ST506 drive)
	public static final int XEBEC_ST412 = 2;	// 10.6MB
	public static final int XEBEC_CM5206 = 3;	//  4.4MB
	public static final int XEBEC_CM5410 = 4;	//  8.9MB
	public static final int XEBEC_CM5616 = 5;	// 13.3MB
	public static final int XEBEC_RO201 = 6;	//  5.5MB
	public static final int XEBEC_RO202 = 7;	// 11.1MB
	public static final int XEBEC_RO203 = 8;	// 16.7MB
	public static final int XEBEC_RO204 = 9;	// 22.3MB
	public static final int Z67IDE = 10;
	public static final int CF = 11;
	public static final int SDCARD = 12;
	private int[][] params = new int[][] {
		{ 0, 0, 0, 0 },		// /* INVALID=0    */
		{ 153, 4, 128, 64 },	// /*[XEBEC_ST506] */
		{ 306, 4, 128, 64 },	// /*[XEBEC_ST412] */
		{ 256, 2, 256, 256 },	// /*[XEBEC_CM5206]*/
		{ 256, 4, 256, 256 },	// /*[XEBEC_CM5410]*/
		{ 256, 6, 256, 256 },	// /*[XEBEC_CM5616]*/
		{ 321, 2, 132, 0 },	// /*[XEBEC_RO201] */
		{ 321, 4, 132, 0 },	// /*[XEBEC_RO202] */
		{ 321, 6, 132, 0 },	// /*[XEBEC_RO203] */
		{ 321, 8, 132, 0 },	// /*[XEBEC_RO204] */
		{ 2048, 8, 0, 0 },	// /*[Z67IDE] (def 128M) */
		{ 4096, 8, 0, 0 },	// /*[CF] (def 256M) */
		{ 4096, 8, 0, 0 },	// /*[SDCARD] (def 256M) */
	};
	private boolean checkHeader(byte[] buf) {
		int m = 0;
		int e = 0;
		while (buf[e] != '\n' && buf[e] != '\0' && e < buf.length) {
			int p = 0;
			while (e < buf.length && Character.isDigit(buf[e])) {
				p = (p * 10) + (buf[e] - '0');
				++e;
			}
			if (e >= buf.length) {
				// too harsh?
				return false;
			}
			// TODO: removable flag, others?
			// NOTE: removable media requires many more changes.
			switch (Character.toLowerCase(buf[e])) {
			case 'c':
				if (p > 0x0ffff) {
					break;
				}
				m |= 0x01;
				cyls = p & 0x0ffff;
				break;

			case 'h':
				if (p > 0x0f) {
					break;
				}
				m |= 0x02;
				hds = p & 0x0f;
				break;

			case 'z':
				if (p != 128 && p != 256 && p !=512 && p != 1024) {
					break;
				}
				m |= 0x04;
				ssz = p & 0x0fff;
				break;

			case 'p':
				if (p > 0x0ff) {
					break;
				}
				m |= 0x08;
				spt = p & 0x0ff;
				break;

			case 'l':
				if (p > 0x0ff) {
					break;
				}
				m |= 0x10;
				lat = p & 0x0ff;
				break;

			default:
				return false;
			}

			e++;
		}
		return m == 0x1f;
	}
	// END GenericSASIDrive code
	public static String[] disks;

	public static final int RESV_SECS = 12;

	// These are all H8/H89 versions
	static private byte[] boot_jmp = new byte[] {
		(byte)0xC3, (byte)0x80, (byte)0x24
	};
// e.g. od -tx1 -w8 git/MmsCpm3/defsasi3/bin/b320bot.bin | sed -f botfixup.sed
	static private byte[] boot = new byte[] {
(byte)0xc3,(byte)0x85,(byte)0x24,(byte)0x00,(byte)0x00,(byte)0x3b,(byte)0x3b,(byte)0xd1,
(byte)0x21,(byte)0xf0,(byte)0x25,(byte)0xc1,(byte)0x79,(byte)0xfe,(byte)0xc3,(byte)0x28,
(byte)0x03,(byte)0x77,(byte)0x78,(byte)0xb7,(byte)0x31,(byte)0x80,(byte)0x26,(byte)0xd5,
(byte)0xc0,(byte)0x3a,(byte)0x34,(byte)0x20,(byte)0xfe,(byte)0x05,(byte)0x30,(byte)0x04,
(byte)0xd6,(byte)0x03,(byte)0x18,(byte)0x02,(byte)0xd6,(byte)0xa8,(byte)0xe6,(byte)0x03,
(byte)0x0f,(byte)0x0f,(byte)0x0f,(byte)0x32,(byte)0xea,(byte)0x25,(byte)0x3a,(byte)0x8d,
(byte)0x22,(byte)0x32,(byte)0xe9,(byte)0x25,(byte)0x3a,(byte)0x91,(byte)0x22,(byte)0x32,
(byte)0xed,(byte)0x25,(byte)0x3a,(byte)0x50,(byte)0x21,(byte)0x4f,(byte)0x0c,(byte)0xcd,
(byte)0x6c,(byte)0x25,(byte)0xcd,(byte)0x8d,(byte)0x25,(byte)0xcd,(byte)0xcb,(byte)0x25,
(byte)0x3a,(byte)0x83,(byte)0x22,(byte)0xe6,(byte)0xe0,(byte)0xc2,(byte)0xe9,(byte)0x24,
(byte)0x0c,(byte)0x3e,(byte)0x0c,(byte)0x32,(byte)0xe9,(byte)0x25,(byte)0xcd,(byte)0x6c,
(byte)0x25,(byte)0xcd,(byte)0x8d,(byte)0x25,(byte)0x21,(byte)0x85,(byte)0x22,(byte)0x06,
(byte)0x08,(byte)0x1e,(byte)0xc8,(byte)0xcd,(byte)0x95,(byte)0x25,(byte)0xcd,(byte)0xcb,
(byte)0x25,(byte)0x21,(byte)0xf0,(byte)0x25,(byte)0x7e,(byte)0xb7,(byte)0x28,(byte)0x07,
(byte)0xd6,(byte)0x30,(byte)0x21,(byte)0x93,(byte)0x22,(byte)0xbe,(byte)0xd0,(byte)0x21,
(byte)0x94,(byte)0x22,(byte)0x4f,(byte)0x06,(byte)0x00,(byte)0x09,(byte)0x09,(byte)0x09,
(byte)0xe5,(byte)0x21,(byte)0x9a,(byte)0x22,(byte)0x11,(byte)0x15,(byte)0x00,(byte)0x19,
(byte)0x0d,(byte)0xf2,(byte)0x07,(byte)0x25,(byte)0x11,(byte)0x0f,(byte)0x00,(byte)0x19,
(byte)0x7e,(byte)0xe1,(byte)0x4e,(byte)0x23,(byte)0x5e,(byte)0x23,(byte)0x56,(byte)0xe6,
(byte)0x03,(byte)0x47,(byte)0x32,(byte)0x83,(byte)0x24,(byte)0x28,(byte)0x13,(byte)0x79,
(byte)0xe6,(byte)0x1f,(byte)0x4f,(byte)0x3e,(byte)0x01,(byte)0xcb,(byte)0x19,(byte)0xcb,
(byte)0x1b,(byte)0xcb,(byte)0x1a,(byte)0x07,(byte)0xb7,(byte)0x10,(byte)0xf6,(byte)0x32,
(byte)0x84,(byte)0x24,(byte)0x3a,(byte)0xea,(byte)0x25,(byte)0xb1,(byte)0x32,(byte)0xea,
(byte)0x25,(byte)0xed,(byte)0x53,(byte)0xeb,(byte)0x25,(byte)0x3e,(byte)0x01,(byte)0x32,
(byte)0xed,(byte)0x25,(byte)0x3a,(byte)0x84,(byte)0x22,(byte)0x32,(byte)0xee,(byte)0x25,
(byte)0x3e,(byte)0x08,(byte)0x32,(byte)0xe9,(byte)0x25,(byte)0x3a,(byte)0x50,(byte)0x21,
(byte)0x4f,(byte)0x0c,(byte)0xcd,(byte)0x6c,(byte)0x25,(byte)0xcd,(byte)0x8d,(byte)0x25,
(byte)0xcd,(byte)0xae,(byte)0x25,(byte)0xcd,(byte)0xcb,(byte)0x25,(byte)0x21,(byte)0xea,
(byte)0x25,(byte)0x11,(byte)0x77,(byte)0x23,(byte)0x01,(byte)0x05,(byte)0x00,(byte)0xed,
(byte)0xb0,(byte)0xc3,(byte)0x80,(byte)0x22,(byte)0x06,(byte)0x00,(byte)0xed,(byte)0x78,
(byte)0xe6,(byte)0x08,(byte)0x28,(byte)0x04,(byte)0x10,(byte)0xf8,(byte)0x18,(byte)0x34,
(byte)0x3e,(byte)0x40,(byte)0xed,(byte)0x79,(byte)0x06,(byte)0x00,(byte)0xed,(byte)0x78,
(byte)0xe6,(byte)0x08,(byte)0x20,(byte)0x04,(byte)0x10,(byte)0xf8,(byte)0x18,(byte)0x24,
(byte)0x3e,(byte)0x00,(byte)0xed,(byte)0x79,(byte)0xc9,(byte)0x0d,(byte)0x21,(byte)0xe9,
(byte)0x25,(byte)0x06,(byte)0x06,(byte)0x1e,(byte)0xd8,(byte)0xc5,(byte)0x0c,(byte)0x06,
(byte)0x00,(byte)0xed,(byte)0x78,(byte)0xe6,(byte)0xd8,(byte)0xbb,(byte)0x28,(byte)0x05,
(byte)0x10,(byte)0xf7,(byte)0xc1,(byte)0x18,(byte)0x07,(byte)0xc1,(byte)0xed,(byte)0xa3,
(byte)0xc2,(byte)0x95,(byte)0x25,(byte)0xc9,(byte)0xd1,(byte)0xc9,(byte)0x21,(byte)0x80,
(byte)0x22,(byte)0x0c,(byte)0xed,(byte)0x78,(byte)0x0d,(byte)0xe6,(byte)0xd8,(byte)0xfe,
(byte)0x98,(byte)0xc8,(byte)0xe6,(byte)0x98,(byte)0xfe,(byte)0x88,(byte)0x20,(byte)0xf1,
(byte)0x3a,(byte)0x84,(byte)0x24,(byte)0x06,(byte)0x80,(byte)0xed,(byte)0xb2,(byte)0x3d,
(byte)0x20,(byte)0xf9,(byte)0xc9,(byte)0x21,(byte)0xef,(byte)0x25,(byte)0x18,(byte)0x03,
(byte)0xed,(byte)0x78,(byte)0x77,(byte)0x0c,(byte)0xed,(byte)0x78,(byte)0x0d,(byte)0xe6,
(byte)0xf0,(byte)0xfe,(byte)0x90,(byte)0x28,(byte)0xf3,(byte)0xfe,(byte)0xb0,(byte)0x20,
(byte)0xf2,(byte)0xed,(byte)0x78,(byte)0x7e,(byte)0xe6,(byte)0x03,(byte)0x20,(byte)0xc4,
(byte)0xc9,(byte)0x00,(byte)0x00,(byte)0x00,(byte)0x00,(byte)0x00,(byte)0x00,(byte)0x00,
(byte)0x00,
	};
	static private byte[] bootc = new byte[] {	// CF-Card boot
(byte)0xc3,(byte)0x85,(byte)0x24,(byte)0x00,(byte)0x00,(byte)0xd1,(byte)0x31,(byte)0x80,
(byte)0x26,(byte)0xd5,(byte)0x3a,(byte)0x34,(byte)0x20,(byte)0xd6,(byte)0x46,(byte)0x21,
(byte)0x93,(byte)0x22,(byte)0xbe,(byte)0xd0,(byte)0x21,(byte)0x94,(byte)0x22,(byte)0x4f,
(byte)0x06,(byte)0x00,(byte)0x09,(byte)0x09,(byte)0x09,(byte)0x7e,(byte)0xe6,(byte)0x1f,
(byte)0x4f,(byte)0x23,(byte)0x5e,(byte)0x23,(byte)0x56,(byte)0xcb,(byte)0x39,(byte)0xcb,
(byte)0x1b,(byte)0xcb,(byte)0x1a,(byte)0xcb,(byte)0x39,(byte)0xcb,(byte)0x1b,(byte)0xcb,
(byte)0x1a,(byte)0x3a,(byte)0x31,(byte)0x21,(byte)0x3c,(byte)0xd3,(byte)0x80,(byte)0x3a,
(byte)0x56,(byte)0x21,(byte)0x32,(byte)0x10,(byte)0x25,(byte)0xf6,(byte)0xe0,(byte)0xd3,
(byte)0x8e,(byte)0x3a,(byte)0x57,(byte)0x21,(byte)0xb1,(byte)0x32,(byte)0x11,(byte)0x25,
(byte)0xd3,(byte)0x8d,(byte)0x7b,(byte)0x32,(byte)0x12,(byte)0x25,(byte)0xd3,(byte)0x8c,
(byte)0x7a,(byte)0x32,(byte)0x13,(byte)0x25,(byte)0xd3,(byte)0x8b,(byte)0x3e,(byte)0x01,
(byte)0xd3,(byte)0x8a,(byte)0x3e,(byte)0x20,(byte)0xd3,(byte)0x8f,(byte)0xdb,(byte)0x8f,
(byte)0xcb,(byte)0x7f,(byte)0x20,(byte)0xfa,(byte)0xcb,(byte)0x47,(byte)0x20,(byte)0x24,
(byte)0xcb,(byte)0x77,(byte)0x28,(byte)0x20,(byte)0xcb,(byte)0x5f,(byte)0x28,(byte)0xee,
(byte)0x0e,(byte)0x88,(byte)0x06,(byte)0x00,(byte)0x21,(byte)0x80,(byte)0x22,(byte)0xed,
(byte)0xb2,(byte)0xed,(byte)0xb2,(byte)0xaf,(byte)0xd3,(byte)0x80,(byte)0x21,(byte)0x10,
(byte)0x25,(byte)0x11,(byte)0x77,(byte)0x23,(byte)0x01,(byte)0x04,(byte)0x00,(byte)0xed,
(byte)0xb0,(byte)0xc3,(byte)0x80,(byte)0x22,(byte)0xaf,(byte)0xd3,(byte)0x80,(byte)0xc9,
(byte)0x00,(byte)0x00,(byte)0x00,(byte)0x00,
	};
	static private byte[] boots = new byte[] {	// SDCARD boot
(byte)0xc3,(byte)0x85,(byte)0x24,(byte)0x00,(byte)0x00,(byte)0xd1,(byte)0x31,(byte)0x80,
(byte)0x26,(byte)0xd5,(byte)0x3a,(byte)0x31,(byte)0x21,(byte)0x3c,(byte)0x07,(byte)0x07,
(byte)0x32,(byte)0xf9,(byte)0x24,(byte)0x3a,(byte)0x34,(byte)0x20,(byte)0xd6,(byte)0x50,
(byte)0x21,(byte)0x93,(byte)0x22,(byte)0xbe,(byte)0xd0,(byte)0x21,(byte)0x94,(byte)0x22,
(byte)0x4f,(byte)0x06,(byte)0x00,(byte)0x09,(byte)0x09,(byte)0x09,(byte)0x7e,(byte)0xe6,
(byte)0x1f,(byte)0x4f,(byte)0x23,(byte)0x5e,(byte)0x23,(byte)0x56,(byte)0xcb,(byte)0x39,
(byte)0xcb,(byte)0x1b,(byte)0xcb,(byte)0x1a,(byte)0xcb,(byte)0x39,(byte)0xcb,(byte)0x1b,
(byte)0xcb,(byte)0x1a,(byte)0x2a,(byte)0x56,(byte)0x21,(byte)0x7c,(byte)0xb1,(byte)0x67,
(byte)0x22,(byte)0xfb,(byte)0x24,(byte)0xeb,(byte)0x22,(byte)0xfd,(byte)0x24,(byte)0x21,
(byte)0x80,(byte)0x22,(byte)0xcd,(byte)0xdc,(byte)0x24,(byte)0xd8,(byte)0x21,(byte)0xfb,
(byte)0x24,(byte)0x11,(byte)0x77,(byte)0x23,(byte)0x01,(byte)0x04,(byte)0x00,(byte)0xed,
(byte)0xb0,(byte)0xc3,(byte)0x80,(byte)0x22,(byte)0xe5,(byte)0x21,(byte)0xfa,(byte)0x24,
(byte)0x16,(byte)0x01,(byte)0x1e,(byte)0x00,(byte)0xcd,(byte)0x01,(byte)0x25,(byte)0xe1,
(byte)0x38,(byte)0x0a,(byte)0x3a,(byte)0x00,(byte)0x25,(byte)0xb7,(byte)0x20,(byte)0x04,
(byte)0xcd,(byte)0x43,(byte)0x25,(byte)0xc9,(byte)0xaf,(byte)0xd3,(byte)0x41,(byte)0x37,
(byte)0xc9,(byte)0x04,(byte)0x51,(byte)0x00,(byte)0x00,(byte)0x00,(byte)0x00,(byte)0x01,
(byte)0x00,(byte)0x3a,(byte)0xf9,(byte)0x24,(byte)0xd3,(byte)0x41,(byte)0x0e,(byte)0x40,
(byte)0xe5,(byte)0x21,(byte)0x00,(byte)0x01,(byte)0xed,(byte)0x78,(byte)0xfe,(byte)0xff,
(byte)0x28,(byte)0x08,(byte)0x2b,(byte)0x7c,(byte)0xb5,(byte)0x20,(byte)0xf5,(byte)0xe1,
(byte)0x37,(byte)0xc9,(byte)0xe1,(byte)0x06,(byte)0x06,(byte)0xed,(byte)0xb3,(byte)0xed,
(byte)0x78,(byte)0xe5,(byte)0x21,(byte)0x00,(byte)0x01,(byte)0xed,(byte)0x78,(byte)0xfe,
(byte)0xff,(byte)0x20,(byte)0x08,(byte)0x2b,(byte)0x7c,(byte)0xb5,(byte)0x20,(byte)0xf5,
(byte)0xe1,(byte)0x37,(byte)0xc9,(byte)0xe1,(byte)0x42,(byte)0x77,(byte)0x23,(byte)0x05,
(byte)0x28,(byte)0x02,(byte)0xed,(byte)0xb2,(byte)0x7b,(byte)0xb7,(byte)0xc8,(byte)0xaf,
(byte)0xd3,(byte)0x41,(byte)0xc9,(byte)0x3a,(byte)0xf9,(byte)0x24,(byte)0xd3,(byte)0x41,
(byte)0x0e,(byte)0x40,(byte)0x11,(byte)0x00,(byte)0x01,(byte)0xed,(byte)0x78,(byte)0xfe,
(byte)0xff,(byte)0x20,(byte)0x08,(byte)0x1b,(byte)0x7a,(byte)0xb3,(byte)0x20,(byte)0xf5,
(byte)0x37,(byte)0x18,(byte)0x10,(byte)0xfe,(byte)0xfe,(byte)0x37,(byte)0x20,(byte)0x0b,
(byte)0x06,(byte)0x00,(byte)0xed,(byte)0xb2,(byte)0xed,(byte)0xb2,(byte)0xed,(byte)0x78,
(byte)0xed,(byte)0x78,(byte)0xaf,(byte)0xf5,(byte)0xaf,(byte)0xd3,(byte)0x41,(byte)0xf1,
(byte)0xc9,
// 0000f1

	};

	static private byte[] idat = new byte[RESV_SECS * 128];

	static {
		Arrays.fill(idat, (byte)0);
		System.arraycopy(boot_jmp, 0, idat, 0, boot_jmp.length);
		disks = new String[16];
		disks[XEBEC_ST506] = "XEBEC_ST506";
		disks[XEBEC_ST412] = "XEBEC_ST412";
		disks[XEBEC_CM5206] = "XEBEC_CM5206";
		disks[XEBEC_CM5410] = "XEBEC_CM5410";
		disks[XEBEC_CM5616] = "XEBEC_CM5616";
		disks[XEBEC_RO201] = "XEBEC_RO201";
		disks[XEBEC_RO202] = "XEBEC_RO202";
		disks[XEBEC_RO203] = "XEBEC_RO203";
		disks[XEBEC_RO204] = "XEBEC_RO204";
		disks[Z67IDE] = "Z67IDE segment";
		disks[CF] = "CF segment";
		disks[SDCARD] = "SDCARD segment";
	}

	int disk = 0;;
	int minBsh; // hint... minimum
	int maxDsm; // hard limit, based on BIOS ALV size
	int spt;
	int ssz;
	int cyls;
	int hds;
	public int resv_trks;
	int resv_secs; // 128-byte sectors
	int lat;
	long capacity; // 'ssz'-byte sectors
	long logCap; // logical capacity (Z67-IDE 128M segment)
	float factor; // padding for directory
	long[] parts = null;
	int nparts;
	byte[] sec0;
	CpmDpb[] dpbs = null;
	// These are transient... setcpmdpb() sets.
	int blocksize;
	int maxdir;
	boolean z67ide;
	boolean cf;
	boolean sdcard;
	long dskBase;

	public H89SASIFormatter(File fi) {
		sec0 = new byte[idat.length];
		byte[] buf = new byte[128];
		try {
			RandomAccessFile fd = new RandomAccessFile(fi, "r");
			fd.read(sec0);
			fd.seek(fd.length() - buf.length);
			fd.read(buf);
			fd.close();
		} catch (Exception ee) {
			return;
		}
		if (!checkHeader(buf)) {
			return;
		}
		try {
			for (int dsk = 1; true; ++dsk) {
				// hopefully these are unique...
				if (params[dsk][0] == cyls && params[dsk][1] == hds) {
					disk = dsk;
					break;
				}
			}
		} catch (Exception ee) {}
		if (disk == 0) {
			// No obvious match to format... but still have
			// partition data.
			System.err.format("Unable to guess at original format type\n");
			disk = CF; // need better guess/punt.
		}
		dskBase = 0;
		z67ide = (disk == Z67IDE);
		cf = (disk == CF);
		sdcard = (disk == SDCARD);
		if (sdcard) {
			System.arraycopy(boots, 0, idat, 512, boots.length);
		} else if (cf) {
			System.arraycopy(bootc, 0, idat, 512, bootc.length);
		} else {
			System.arraycopy(boot, 0, idat, 512, boot.length);
		}
		// TODO: actual capacity is 128M...
		capacity = cyls * hds * spt;
		if (z67ide) {
			logCap = (128*1024*1024) / ssz;
			// TODO: alter base offset...
			// dskBase = (segment << 27);	// 2^27 = 128M
		} else if (cf || sdcard) {
			logCap = (256*1024*1024) / ssz;
			// TODO: alter base offset...
			// dskBase = (segment << 28);	// 2^28 = 256M
		} else {
			logCap = capacity;
		}
		resv_secs = RESV_SECS; // per disk, for init data. 128-byte sectors!
		resv_trks = 2; // each partition, for CP/M...
		parts = new long[8]; // allocate all 8 anyway
		Arrays.fill(parts, 0);
		dpbs = new CpmDpb[8];
		int adr = 19;
		int adr2 = 47;
		nparts = sec0[adr++] & 0xff;
		int t = resv_secs;
		int sxf = ssz / 128;
		// convert partition table from "start sector"
		// to "length in sectors".
		for (int x = 0; x < nparts; ++x) {
			// this is the partition start address...
			parts[x] = getpart(sec0, adr);
			if (x > 0) {
				parts[x - 1] = parts[x] - parts[x - 1];
				t += parts[x - 1];
			}
			dpbs[x] = new CpmDpb();
			getcpmdpb(sec0, adr2, dpbs[x]);
			dpbs[x].poff = t;
			adr += 3;;
			adr2 += 21;;
		}
		// This over-inflates the last partition sometimes,
		// the DPB tells the true CP/M capacity - but that's
		// not the whole story, either.
		parts[nparts - 1] = capacity * sxf - parts[nparts - 1];
	}

	public H89SASIFormatter(int dsk, String parm, File boot0,
			int sz, int lt, int mBsh, int mDsm, float fact) {
		dskBase = 0;
		z67ide = (dsk == Z67IDE);
		cf = (dsk == CF);
		sdcard = (dsk == SDCARD);
		if (boot0 != null) {
			setupBootImage(boot0, idat);
		} else if (sdcard) {
			System.arraycopy(boots, 0, idat, 512, boots.length);
		} else if (cf) {
			System.arraycopy(bootc, 0, idat, 512, bootc.length);
		} else {
			System.arraycopy(boot, 0, idat, 512, boot.length);
		}
		if (z67ide || cf || sdcard) {
			// assumes CF Card, but general IDE/ATA also works.
			// NOTE: Z67IDE isolates user from CF card sector size...
			ssz = 512;
			spt = 16;
			if (z67ide) {
				ssz = sz;
				if (ssz == 256) {
					spt = 32;
				}
			}
			if (parm == null) {
				capacity = (z67ide ? 128 : 256) * 1024 * 1024;
			} else {
				String[] ss = parm.split("_");
				if (ss.length != 2) {
					return;
				}
				if (!ss[1].matches("[0-9]+[MG]")) {
					return;
				}
				int i = ss[1].length() - 1;
				char f = ss[1].charAt(i);
				capacity = Integer.valueOf(ss[1].substring(0, i)) *
						1024 * 1024;
				if (f == 'G') {
					capacity *= 1024;
				}
			}
			capacity /= ssz;
			if (capacity < 1024*1024*1024/ssz) { // < 1G
				hds = 8;
				cyls = (int)(capacity / spt / hds);
			} else {
				cyls = 16384;
				hds = (int)(capacity / spt / cyls);
			}
			params[dsk][0] = cyls;
			params[dsk][1] = hds;
		} else {
			// Standard, legacy, SASI drives
			if (sz == 512) {
				spt = 17;
			} else if (sz == 256) {
				spt = 32;
			} else {
				return;
			}
			try {
				cyls = params[dsk][0];
				hds = params[dsk][1];
			} catch (Exception ee) {
				// likely "array index out of bounds"...
				return;
			}
			ssz = sz;
			capacity = cyls * hds * spt;
		}
		disk = dsk;
		resv_secs = RESV_SECS; // per disk, for init data. 128-byte sectors!
		resv_trks = 2; // each partition, for CP/M...
		lat = lt; // not used?
		minBsh = mBsh; // minimum value
		maxDsm = mDsm; // maximum value
		factor = fact;
	}

	// This must be called before partitions are setup in 'img'.
	// This will overwrite all of 'img'.
	private void setupBootImage(File boot0, byte[] img) {
		// TODO: errors
		try {
			InputStream f = new FileInputStream(boot0);
			f.read(img);
			f.close();
		} catch (Exception ee) {
			ee.printStackTrace();
		}
	}

	public void setSegment(int seg) {
		if (z67ide) {
			dskBase = (seg << 27);	// 2^27 = 128M
		}
	}

	public void dump(PrintStream out) {
		double cap = getCapacity();
		cap /= 1024;
		cap /= 1024;
		out.format("Drive %s (%.1f MB)\n", disks[disk], cap);
		out.format("Sector Size %d, Sec/Trk %d, Cylinders %d, Heads %d\n",
			ssz, spt, cyls, hds);
		for (int x = 0; x < nparts; ++x) {
			cap = parts[x] * 128;
			cap /= 1024;
			cap /= 1024;
			out.format("Partition %d %dx%d (%.1fMB) CP/M DPB:\n", x,
				parts[x], ssz, cap);
			out.format("    %d size\n",
				((dpbs[x].dsm + 1) << dpbs[x].bsh) * 128);
			out.format("    %d spt\n", dpbs[x].spt);
			out.format("    %d,%d,%d bsh,blm,exm\n",
				dpbs[x].bsh, dpbs[x].blm, dpbs[x].exm);
			out.format("    %d,%d dsm,drm\n",
				dpbs[x].dsm, dpbs[x].drm);
			out.format("    %s alv0\n", Integer.toBinaryString(dpbs[x].al0));
			out.format("    %d,%d cks,off\n",
				dpbs[x].cks, dpbs[x].off);
			out.format("    %d [poff]\n", dpbs[x].poff);
		}
	}

	// Total capacity in bytes
	public long getCapacity() { return capacity * ssz; }

	// Available number of sectors (ssz).
	public long getAvailable() { return capacity - (resv_secs * 128) / ssz; }

	// Partition Table (pt[]) is in 'sz' size sectors (not 128-byte).
	// ddfi may be null.
	public boolean setPartitions(long[] pt, int np, File ddfi, CpmDpb[] dpbs) {
		if (isError() || np <= 0 || np > 8) {
			return false;
		}
		String basename = "";
		PrintStream ddfp = null;
		if (ddfi != null) {
			basename = ddfi.getName();
			try {
				ddfp = new PrintStream(ddfi);
			} catch (Exception ee) {}
		}
		sec0 = Arrays.copyOf(idat, idat.length);
		int adr = 3;
		Arrays.fill(sec0, adr, 256, (byte)0);
		sec0[adr++] = (byte)0b00011111;	// drive/ctrlr code TODO: real codes...
		sec0[adr++] = (byte)1;		// control byte
		setdchars(sec0, adr, disk); // 8 bytes...
		adr += 8;
		// assign drive type - not used. 6 bytes.
		adr += 6;
		sec0[adr++] = (byte)np;
		long t = resv_secs;
		int sxf = ssz / 128;
		for (int x = 0; x < 9; ++x) {
			if (x < np) {
				System.out.format("Partition %d start %d\n", x, t);
				setpart(sec0, adr, t);
				t += pt[x] * sxf;
			} else {
				setpart(sec0, adr, 0);
			}
			adr += 3;
		}
		// +47 bytes. Now CP/M DPBs for each partition...  21 bytes each
		t = resv_secs;
		for (int x = 0; x < np; ++x) {
			setcpmdpb(sec0, adr, pt[x], dpbs[x]);
			adr += 21;
			if (ddfp != null) {
				int lspt = 8192 / ssz;
				ddfp.format("diskdef %s.%d.%d\n" +
					"  seclen %d\n  tracks %d\n  sectrk %d\n" +
					"  blocksize %d\n  maxdir %d\n  skew 0\n" +
					"  boottrk 2\n  offset %d\n  os 2.2\nend\n",
					basename, np, x,
					ssz, pt[x] / lspt, lspt,
					blocksize, maxdir,
					t * 128);
			}
			t += pt[x];
		}
		return true;
	}

	public boolean isError() { return disk == 0; }

	private void setdb(byte[] buf, int start, int db) {
		buf[start] = (byte)(db & 0x0ff);
	}

	private void setdw(byte[] buf, int start, int dw) {
		// little-endian
		buf[start++] = (byte)(dw & 0x0ff);
		buf[start++] = (byte)((dw >> 8) & 0x0ff);
	}

	private void setd2(byte[] buf, int start, int dw) {
		// big-endian
		buf[start++] = (byte)((dw >> 8) & 0x0ff);
		buf[start++] = (byte)(dw & 0x0ff);
	}

	private void setdchars(byte[] buf, int start, int disk) {
		setd2(buf, start + 0, params[disk][0]);
		setdb(buf, start + 2, params[disk][1]);
		setd2(buf, start + 3, params[disk][2]);
		setd2(buf, start + 5, params[disk][3]);
		buf[start + 7] = (byte)0; // ECC burst length - not relavent
	}

	private int getpart(byte[] buf, int start) {
		// big-endian
		int p = (buf[start++] & 0xff) << 16;
		p |= (buf[start++] & 0xff) << 8;
		p |= buf[start++] & 0xff;
		return p;
	}

	private void setpart(byte[] buf, int start, long pstart) {
		// big-endian
		buf[start++] = (byte)((pstart >> 16) & 0x0ff);
		buf[start++] = (byte)((pstart >> 8) & 0x0ff);
		buf[start++] = (byte)(pstart & 0x0ff);
	}

	private void getcpmdpb(byte[] buf, int start, CpmDpb dpb) {
		dpb.spt = buf[start++] & 0xff;
		dpb.spt |= (buf[start++] & 0xff) << 8;
		dpb.bsh = buf[start++] & 0xff;
		dpb.blm = buf[start++] & 0xff;
		dpb.exm = buf[start++] & 0xff;
		dpb.dsm = buf[start++] & 0xff;
		dpb.dsm |= (buf[start++] & 0xff) << 8;
		dpb.drm = buf[start++] & 0xff;
		dpb.drm |= (buf[start++] & 0xff) << 8;
		dpb.al0 = (buf[start++] & 0xff) << 8; // big-endian
		dpb.al0 |= buf[start++] & 0xff;
		dpb.cks = buf[start++] & 0xff;
		dpb.cks |= (buf[start++] & 0xff) << 8;
		dpb.off = buf[start++] & 0xff;
		dpb.off |= (buf[start++] & 0xff) << 8;
		// caller must set poff...
	}

	private int setcpmdpb(byte[] buf, int start, long psize, CpmDpb dpb) {
		blocksize = (1 << dpb.bsh) * 128;
		maxdir = dpb.drm + 1;
		setdw(buf, start + 0, 64); // SPT - constant, regardless of hardware.
		setdb(buf, start + 2, dpb.bsh);
		setdb(buf, start + 3, dpb.blm);
		setdb(buf, start + 4, dpb.exm);
		setdw(buf, start + 5, dpb.dsm);
		setdw(buf, start + 7, dpb.drm);
		setd2(buf, start + 9, dpb.al0);	// essentially big-endian...
		setdw(buf, start + 11, 0);	// fixed media, no CKS
		setdw(buf, start + 13, dpb.off);
		// non-CP/M:
		int szc = ssz >> 8;
		if (szc > 2) szc = 3;
		setdb(buf, start + 15, szc);	// sector size code...
		setdb(buf, start + 16, 0b10000000);
		setdb(buf, start + 17, 0b00000000);
		setdb(buf, start + 18, 0b11111111);
		setdb(buf, start + 19, 0b11111111);
		setdb(buf, start + 20, 0b11111111);
		return 0;
	}

	public boolean convert(File fi, File fo, int ssz) {
		// Like format() but convert to new sector size
		// TODO: implement this
		if (ssz != 512 && ssz != 256) {
			return false;
		}
		boolean shrink = ssz < this.ssz;
		boolean grow = ssz > this.ssz;
		this.ssz = ssz;	// sanity check this...
		spt = ssz == 512 ? 17 : 32;
		capacity = cyls * hds * spt;
		int sxf = ssz / 128;
		// TODO: truncate last partition/dpb when smaller ssz...
		// (expand when larger?)
		// essentially, re-compute last partition size
		int tot = 0;
		int x = 0;
		for (; x < nparts - 1; ++x) {
			tot += parts[x];
		}
		// x < nparts (x == nparts - 1)
		parts[x] = capacity * sxf - tot;
		// recompute DPB DSM... this might totally bork it...
		// but it is always safe to use smaller size...
		// (except if >255 changes to <256?)
		int dsm = (int)((parts[x] >> dpbs[x].bsh) - 1);
		if (dsm < dpbs[x].dsm) {
			dpbs[x].dsm = dsm;
		}
		setPartitions(parts, nparts, null, dpbs);
		try {
			InputStream fe = new FileInputStream(fi);
			OutputStream fd = new FileOutputStream(fo);
			// TODO: seek(dskBase)... multiple segments...
			fd.write(sec0);

			byte[] buf = new byte[ssz];
			// skip disk header (already processed)
			// TODO: seek(dskBase)...
			for (x = 0; x < sec0.length / ssz; ++x) {
				if (fe.read(buf) < ssz) {
					return false;	// close files?
				}
			}
			for (; x < capacity; ++x) {
				if (fe.read(buf) < ssz) {
					Arrays.fill(buf, (byte)0xe5);
				}
				fd.write(buf);
			}
			setHeader(fd, buf);
			fd.close();
			return true;
		} catch (Exception ee) {
			ee.printStackTrace();
			return false;
		}
	}

	private byte sdCrc = 0;
	private byte sdCrc7(int b, int n) {
		byte r = sdCrc;
		byte c = (byte)b;
		for (int x = 0; x < n; ++x) {
			r <<= 1;
			if ((c & 0x80) != 0) r += 1;
			if ((r & 0x80) != 0) r ^= 0x09;
			c <<= 1;
		}
		sdCrc = (byte)(r & 0x7f);
		return (byte)b;
	}

	// output file position must be end of data...
	private void setHeader(OutputStream fd, byte[] buf) throws Exception {
		Arrays.fill(buf, (byte)0);
		String hdr = String.format("%dc%dh%dz%dp%dl\n",
			cyls, hds, ssz, spt, lat);
		byte[] b = hdr.getBytes();
		System.arraycopy(b, 0, buf, 0, b.length);
		if (sdcard) {
			// setup CID...
			sdCrc = 0;
			int x = 128 - 16;	// last 16 bytes is CID
			buf[x++] = sdCrc7(0, 8);	// MID - what value is "local"?
			buf[x++] = sdCrc7('Z', 8);	// OID - "ZZ"
			buf[x++] = sdCrc7('Z', 8);	// ...
			buf[x++] = sdCrc7('M', 8);	// PNM - "MMS89"
			buf[x++] = sdCrc7('M', 8);	// ...
			buf[x++] = sdCrc7('S', 8);	// ...
			buf[x++] = sdCrc7('8', 8);	// ...
			buf[x++] = sdCrc7('9', 8);	// ...
			buf[x++] = sdCrc7(0x10, 8);	// PRV - 1.0
			int sn = (int)(new Date().getTime() / 1000);
			buf[x++] = sdCrc7(sn >> 24, 8);	// PSN - serial num
			buf[x++] = sdCrc7(sn >> 16, 8);	// ...
			buf[x++] = sdCrc7(sn >> 8, 8);	// ...
			buf[x++] = sdCrc7(sn, 8);	// ...
			Calendar cal = Calendar.getInstance();
			int y = cal.get(Calendar.YEAR) - 2000;
			int m = cal.get(Calendar.MONTH) + 1;
			buf[x++] = sdCrc7(y >> 4, 8);	// MDT - mfg date
			buf[x++] = sdCrc7(((y & 0x0f) << 4) | m, 8); // ...
			sdCrc7(0, 7);
			buf[x++] = (byte)((sdCrc << 1) | 1);	// CRC, stop bit
			// setup CSD... hard-coded to SDC v2...
			x = 128 - 32;
			// TODO: overflow possible? (22 bits max, 0x03fffff)
			int cap = (int)((capacity >> 10) - 1);
			sdCrc = 0;
			buf[x++] = sdCrc7(0x40, 8);	// [0] v2
			buf[x++] = sdCrc7(0x0e, 8);	// [1] TAAC
			buf[x++] = sdCrc7(0, 8);	// [2] NSAC
			buf[x++] = sdCrc7(0x5a, 8);	// [3] TRAN_SPEED
			buf[x++] = sdCrc7(0x5b, 8);	// [4] CCC(hi)
			buf[x++] = sdCrc7(0x59, 8);	// [5] CCC(lo),READ_BL_LEN
			buf[x++] = sdCrc7(0, 8);	// [6] resv...
			buf[x++] = sdCrc7(cap >> 16, 8); // [7] C_SIZE
			buf[x++] = sdCrc7(cap >> 8, 8);	// [8] ...
			buf[x++] = sdCrc7(cap, 8);	// [9] ...
			buf[x++] = sdCrc7(0x7f, 8);	// [10] ERASE_BLK_EN,SECTOR_SIZE...
			buf[x++] = sdCrc7(0x80, 8);	// [11] ...SECTOR_SIZE,WP_GRP_SIZE
			buf[x++] = sdCrc7(0x0a, 8);	// [12] WP_GRP_EN..
			buf[x++] = sdCrc7(0x40, 8);	// [13] ...
			buf[x++] = sdCrc7(0, 8);	// [14] ...
			sdCrc7(0, 7);
			buf[x++] = (byte)((sdCrc << 1) | 1); // [15] CRC, stop bit
			
		}
		fd.write(buf, 0, 128);
	}

	public boolean format(File fi, boolean repair) {
		// TODO: handle truncate vs. overwrite... need RandomAccessFile?
		if (sec0 == null) {
			sec0 = Arrays.copyOf(idat, idat.length);
			Arrays.fill(sec0, (byte)0xe5);
		}
		try {
			// TODO: RandomAccessFile... for seek()
			OutputStream fd = new FileOutputStream(fi);
			// TODO: seek(dskBase)... do after format?
			fd.write(sec0);

			if (!repair) {
				byte[] buf = new byte[ssz];
				Arrays.fill(buf, (byte)0xe5);
				for (int x = sec0.length / ssz; x < capacity; ++x) {
					fd.write(buf);
				}
				setHeader(fd, buf);
			}
			fd.close();
			return true;
		} catch (Exception ee) {
			ee.printStackTrace();
			return false;
		}
	}
}
