// Copyright (c) 2016 Douglas Miller <durgadas311@gmail.com>

import java.io.FileInputStream;
import java.util.Properties;
import java.io.*;

public class H88Memory extends H89Roms implements Memory, GppListener {
	// One single bank of 48K, plus ROMs... But RAM is fixed at 2000-E000...
	private byte[] mem;
	private int top;
	private int bot;
	private int h89_gppOrg0;
	private int top8k;

	public H88Memory(Properties props, GeneralPurposePort gpio, Interruptor intr) {
		super(props, gpio, intr);	// cannot be ROMX
		int jj501_2 = 0;
		String s = props.getProperty("jj501");
		if (s != null) {
			try {
				int n = Integer.valueOf(s);
				if (n == 1) {
					jj501_2 |= 1;
				}
			} catch (Exception ee) {}
		}
		s = props.getProperty("jj502");
		if (s != null) {
			try {
				int n = Integer.valueOf(s);
				if (n == 1) {
					jj501_2 |= 2;
				}
			} catch (Exception ee) {}
		}
		int memsiz = 16384 << jj501_2;
		if (memsiz > 49152) {
			memsiz = 49152;
		}
		top8k = memsiz;
		mem = new byte[memsiz + 8192];	// at 0x0000
		reset();
		s = props.getProperty("u517");
		if (s != null && s.equalsIgnoreCase("org0")) {
			h89_gppOrg0 = gpio.getOrg0Bit();
			gpio.addGppListener(this);
		}
	}

	public int read(boolean rom, int bank, int address) {
		address &= 0xffff; // necessary?
		if (address < bot) {
			// TODO: does this relocate 8K at 0000?
			if (address < monMask) {
				return mon.read(address);
			}
			if (address < 0x1400) {
				return 0;
			}
			if (address >= 0x1800 && h17 != null) {
				return h17.read(address & 0x07ff);
			}
		} else if (address >= top) {
			return 0;
		}
		return mem[address] & 0xff;
	}
	public int read(int address) {
		return read(true, 0, address);
	}
	public void write(int address, int value) {
		address &= 0xffff; // necessary?
		if (address >= top) {
			return;
		}
		mem[address] = (byte)value;
	}

	public void reset() {
		top = mem.length;
		bot = 0x2000;
	}

	public void dumpCore(String file) {
		try {
			OutputStream core = new FileOutputStream(file);
			core.write(mem);
			core.close();
		} catch (Exception ee) {
			H89Operator.error(null, "Core Dump", ee.getMessage());
		}
	}

	public int interestedBits() {
		return h89_gppOrg0;
	}

	public void gppNewValue(int gpio) {
		if ((gpio & h89_gppOrg0) != 0) {
			top = mem.length - 8192;
			bot = 0x0000;
			// simulate moving of top 8K to 0000
			System.arraycopy(mem, top8k, mem, 0x0000, 8192);
		} else {
			// simulate moving of 8K at 0000 back to top 8K
			System.arraycopy(mem, 0x0000, mem, top8k, 8192);
			top = mem.length;
			bot = 0x2000;
		}
	}
}
