// Copyright (c) 2016 Douglas Miller
import java.awt.*;
import javax.swing.*;
import java.io.*;
import java.util.Properties;
import java.lang.reflect.Field;
import java.util.Properties;

public class VirtualH89 {
	private static JFrame front_end;

	public static void main(String[] args) {
		java.net.URL url = VirtualH89.class.getResource("VirtualH89.class");
		String jar = "jar:" + url.getFile().replace("VirtualH89.class", "");
		//System.err.println(jar);
		// jar:file:/path/to/jar/H89jni.jar!/

		Properties props = new H19Properties("89", args);

		boolean isH8 = H89.isH8(props);
		String model = isH8 ? "H8" : "H89";

		CrtScreen screen = new CrtScreen(props);

		front_end = new JFrame("Virtual " + model + " Computer");
		front_end.getContentPane().setName(model + " Emulator");
		front_end.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		front_end.getContentPane().setBackground(new Color(100, 100, 100));
		// This allows TAB to be sent
		front_end.setFocusTraversalKeysEnabled(false);
		String s = props.getProperty("h89_nameplate");
		if (s == null) {
			if (isH8) {
				s = "np-h19.png";
			} else {
				s = "np-heath1.png";
			}
		}
		LEDHandler lh = null;
		if (s.equals("none")) {
			front_end.add(screen);
		} else {
			Object obj = s;
			if (s.matches("np-.*\\.png")) try {
				obj = new ImageIcon(VirtualH89.class.getResource(s));
			} catch (Exception ee) { }
			HxxFrontSide front = new HxxFrontSide(front_end, screen, props, obj);
			front_end.add(front);
			lh = front;
		}

		H89 h89 = new H89(props, lh);
		// All LEDs should be registered now...
		H89Operator op = new H89Operator(front_end, props, screen, lh);
		op.setCommander(h89.getCommander());

		H19Keyboard kbd = new H19Keyboard(h89.getOutputStream(), false);
		front_end.addKeyListener(kbd);
		screen.setPasteListener(kbd);
		H19Terminal term = new H19Terminal(props, screen, kbd, h89.getInputStream());
		kbd.addResetListener(term);
		kbd.addResetListener(op);

		front_end.pack();
		front_end.setVisible(true);

		h89.start(); // spawns its own thread... returns immediately
		term.run(); // consumes caller's thread... i.e. does not return...
	}
}
