// Copyright (c) 2018 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Properties;
import java.io.*;
import java.net.*;
import java.awt.event.*;
import javax.swing.Timer;

public class WIZ810MJ implements IODevice, ActionListener, Runnable {
	private static final int nsocks = 4;

	private static final int MR = 0;
	private static final int GAR = 1;	// x4
	private static final int SUBR = 5;	// x4
	private static final int SHAR = 9;	// x6
	private static final int SIPR = 15;	// x4
	private static final int IR = 21;
	private static final int IMR = 22;
	private static final int RTR = 23;	// x2
	private static final int RCR = 25;
	private static final int RMSR = 26;
	private static final int TMSR = 27;
	private static final int PATR = 28;	// x2
	private static final int PTIMER = 40;
	private static final int PMAGIC = 41;	// used for client ID
	private static final int UIPR = 42;	// x4
	private static final int UPORT = 46;	// x2

	private static final int defRTR = 2000;
	private static final int defRCR = 8;
	private static final int defPTIMER = 40;

	private static final int Sn_MR = 0;
	private static final int Sn_CR = 1;
	private static final int Sn_IR = 2;
	private static final int Sn_SR = 3;
	private static final int Sn_PORT = 4;	// x2
	private static final int Sn_DHAR = 6;	// x6
	private static final int Sn_DIPR = 12;	// x4
	private static final int Sn_DPORT = 16;	// x2
	private static final int Sn_MSSR = 18;	// x2
	private static final int Sn_PROTO = 20;
	private static final int Sn_TOS = 21;
	private static final int Sn_TTL = 22;
	private static final int Sn_TX_FSR = 32; // x2
	private static final int Sn_TX_RD = 34; // x2
	private static final int Sn_TX_WR = 36; // x2
	private static final int Sn_RX_RSR = 38; // x2
	private static final int Sn_RX_RD = 40; // x2
	private static final int Sn_RX_WR = 42; // (implied) x2

	private static final int defTTL = 128;
	private static final int defDHAR = 0xff; // x6

	// Sn_MR values...
	private static final byte Sn_MR_MULTI = (byte)0x80;
	private static final byte Sn_MR_ND_MC = (byte)0x20;
	private static final byte Sn_MR_PROTO = (byte)0x0f;	// mask
	private static final byte Sn_MR_CLOSED = (byte)0x00;
	private static final byte Sn_MR_TCP = (byte)0x01;
	private static final byte Sn_MR_UDP = (byte)0x02;
	private static final byte Sn_MR_IPRAW = (byte)0x03;
	private static final byte Sn_MR_MACRAW = (byte)0x04;
	private static final byte Sn_MR_PPPOE = (byte)0x05;

	// Sn_CR values...
	private static final byte IDLE = (byte)0x00;
	private static final byte OPEN = (byte)0x01;
	private static final byte LISTEN = (byte)0x02;
	private static final byte CONNECT = (byte)0x04;
	private static final byte DISCON = (byte)0x08;
	private static final byte CLOSE = (byte)0x10;
	private static final byte SEND = (byte)0x20;
	private static final byte SEND_MAC = (byte)0x21;
	private static final byte SEND_KEEP = (byte)0x22;
	private static final byte RECV = (byte)0x40;

	// Sn_SR values...
	private static final byte SOCK_CLOSED = (byte)0x00;
	private static final byte SOCK_INIT = (byte)0x13;
	private static final byte SOCK_LISTEN = (byte)0x14;
	private static final byte SOCK_ESTABLISHED = (byte)0x17;
	private static final byte SOCK_CLOSE_WAIT = (byte)0x1c;
	private static final byte SOCK_UDP = (byte)0x22;
	private static final byte SOCK_IPRAW = (byte)0x32;
	private static final byte SOCK_MACRAW = (byte)0x42;
	private static final byte SOCK_PPPOE = (byte)0x5f;
	private static final byte SOCK_SYNSENT = (byte)0x15;
	private static final byte SOCK_SYNRECV = (byte)0x16;
	private static final byte SOCK_FIN_WAIT = (byte)0x18;
	private static final byte SOCK_CLOSING = (byte)0x1a;
	private static final byte SOCK_TIME_WAIT = (byte)0x1b;
	private static final byte SOCK_LAST_ACK = (byte)0x1d;
	private static final byte SOCK_ARP = (byte)0x11; // 0x21, 0x31

	// Sn_IR bits...
	private static final byte Sn_IR_PRECV = (byte)0x80;
	private static final byte Sn_IR_PFAIL = (byte)0x40;
	private static final byte Sn_IR_PNEXT = (byte)0x20;
	private static final byte Sn_IR_SENDOK = (byte)0x10;
	private static final byte Sn_IR_TIMEOUT = (byte)0x08;
	private static final byte Sn_IR_RECV = (byte)0x04;
	private static final byte Sn_IR_DISCON = (byte)0x02;
	private static final byte Sn_IR_CON = (byte)0x01;

	private String name;
	private int base;
	private int irq;
	private int src;
	private Interruptor intr;

	private Socket[] socks;
	private InputStream[] socki;
	private OutputStream[] socko;
	private int[] txbase;
	private int[] txmask;
	private int[] rxbase;
	private int[] rxmask;
	private byte[] mem;
	private byte[] com_regs;
	private byte[] sock_regs;
	private int ints;
	private int adr;
	private java.util.concurrent.LinkedBlockingDeque<Integer> cmd;
	private LED led;
	private javax.swing.Timer timer;

	public WIZ810MJ(Properties props, String id, int port, int irq,
				LEDHandler lh, Interruptor intr) {
		name = id;
		base = port;
		this.intr = intr;
		this.irq = irq;
		ints = 0;
		String s = props.getProperty("wiz810_port");
		if (s != null) {
			base = Integer.decode(s) & 0xff;
		}
		s = props.getProperty("wiz810_intr");
		if (s != null) {
			this.irq = Integer.decode(s) & 0xff;
		}
		src = intr.registerINT(irq);
		cmd = new java.util.concurrent.LinkedBlockingDeque<Integer>();
		socks = new Socket[nsocks];
		socki = new InputStream[nsocks];
		socko = new OutputStream[nsocks];
		txbase = new int[nsocks];
		txmask = new int[nsocks];
		rxbase = new int[nsocks];
		rxmask = new int[nsocks];
		mem = new byte[16384];
		com_regs = new byte[48];
		sock_regs = new byte[1024];
		led = lh.registerLED(getDeviceName(), LED.Colors.AMBER);
		timer = new javax.swing.Timer(500, this);
		reset();
		Thread t = new Thread(this);
		t.start();
		t = new Thread(new Receiver());
		t.start();
		System.err.format("WIZ810MJ Ethernet Module at port %02x\n", base);
	}

	public void reset() {
		for (int x = 0; x < nsocks; ++x) {
			if (socks[x] != null) try {
				socks[x].close();
				socks[x] = null;
			} catch (Exception ee) {
			}
		}
		adr = 0;
		Arrays.fill(com_regs, (byte)0);
		Arrays.fill(sock_regs, (byte)0);
		com_regs[RTR] = (byte)(defRTR >> 8);
		com_regs[RTR+1] = (byte)defRTR;
		com_regs[RCR] = (byte)defRCR;
		com_regs[RMSR] = (byte)0x55;
		com_regs[TMSR] = (byte)0x55;
		com_regs[PTIMER] = (byte)defPTIMER;
		// PTIMER = 40
		for (int x = 0; x < nsocks; ++x) {
			int ix = (x << 8);
			sock_regs[ix + Sn_TTL] = (byte)defTTL;
			Arrays.fill(sock_regs, ix + Sn_DHAR, ix + Sn_DHAR + 6,
								(byte)defDHAR);
		}
		setBuff(com_regs[RMSR], false);
		setBuff(com_regs[TMSR], true);
	}

	public int getBaseAddress() { return base; }
	public int getNumPorts() { return 4; }

	public int in(int port) {
		// TODO: only allow INDIR
		int val = 0;
		int off = port - base;
		switch (off) {
		case 0:
			val = com_regs[MR] & 0xff;
			break;
		case 1:
			val = (adr >> 8) & 0xff;
			break;
		case 2:
			val = adr & 0xff;
			break;
		case 3:
			val = read();
			break;
		}
		return val;
	}
	public void out(int port, int val) {
		// TODO: only allow INDIR
		int off = port - base;
		switch (off) {
		case 0:
			if ((val & 0x80) != 0) {
				// TODO: full reset? or less?
				reset();
			} else {
				com_regs[MR] = (byte)val;
			}
			break;
		case 1:
			adr &= 0x00ff;
			adr |= (val & 0x7f) << 8;
			break;
		case 2:
			adr &= 0x7f00;
			adr |= (val & 0xff);
			break;
		case 3:
			write(val);
			break;
		}
	}

	private boolean ledState = false;
	private void startLED() {
		if (led == null) {
			return;
		}
		synchronized(this) {
			if (!ledState) {
				led.set(true);
				ledState = true;
			}
			timer.removeActionListener(this);
			timer.addActionListener(this);
			timer.restart();
		}
	}

	public void actionPerformed(ActionEvent e) {
		if (e.getSource() == timer) {
			if (led == null) {
				ledState = false;
				return;
			}
			synchronized(this) {
				timer.removeActionListener(this);
				led.set(false);
				ledState = false;
			}
			return;
		}
	}

	private void chkIntr() {
		int i = (com_regs[IR] & com_regs[IMR]);
		if (i == ints) {
			return;
		}
		ints = i;
		if (ints != 0) {
			intr.raiseINT(irq, src);
		} else {
			intr.lowerINT(irq, src);
		}
	}

	private void chkIntr(int ix) {
		int iv = com_regs[IR] & 0xff;
		int im = (1 << (ix >> 8));
		if (sock_regs[ix + Sn_IR] != 0) {
			iv |= im;
		} else {
			iv &= ~im;
		}
		com_regs[IR] = (byte)iv;
		chkIntr();
	}

	private int read() {
		int val = 0;
		if (adr < 0x0030) {
			val = com_regs[adr] & 0xff;
		} else if (adr >= 0x0400 && adr < 0x0800) {
			val = sock_regs[adr & 0x3ff] & 0xff;
		} else if (adr >= 0x4000) {
			val = mem[adr & 0x3fff] & 0xff;
		}
		if ((com_regs[MR] & 0x03) == 0x03) {
			adr = (adr + 1) & 0x7fff;
		}
		return val;
	}

	private void write(int val) {
		if (adr < 0x0030) {
			if (adr == IR) {
				val &= 0xf0;
				com_regs[adr] &= ~val;
				return;
			}
			com_regs[adr] = (byte)val;
			if (adr == RMSR || adr == TMSR) {
				setBuff(val, (adr == TMSR));
			} else if (adr == IMR) {
				chkIntr();
			}
		} else if (adr >= 0x0400 && adr < 0x0800) {
			int ix = (adr & 0x300);
			int of = (adr & 0x0ff);
			if (of == Sn_IR) {
				sock_regs[ix + Sn_IR] &= ~val;
				chkIntr(ix);
				return;
			}
			sock_regs[adr & 0x3ff] = (byte)val;
			if (of == Sn_CR) {
				cmd.add(ix);
			}
		} else if (adr >= 0x4000) {
			mem[adr & 0x3fff] = (byte)val;
		}
		if ((com_regs[MR] & 0x03) == 0x03) {
			adr = (adr + 1) & 0x7fff;
		}
	}

	private void setBuff(int val, boolean tx) {
		int b = tx ? 0x0000 : 0x2000;
		for (int x = 0; x < nsocks; ++x) {
			int ix = (x << 8);
			int m = (1024 << (val & 3));
			if (tx) {
				// LSB is always zero...
				sock_regs[ix + Sn_TX_FSR] = (byte)(m >> 8);
				txbase[x] = b;
				txmask[x] = m - 1;
			} else {
				rxbase[x] = b;
				rxmask[x] = m - 1;
			}
			b += m;
			val >>= 2;
		}
	}

	private void doOpen(int ix) {
		// anything to do here?
		int proto = sock_regs[ix + Sn_MR] & Sn_MR_PROTO;
		if (proto != Sn_MR_TCP) {
System.err.format("WIZ810MJ: unsupported protocol %d\n", proto);
			sock_regs[ix + Sn_SR] = SOCK_CLOSED;
			return;
		}
		sock_regs[ix + Sn_SR] = SOCK_INIT;
		//chkIntr(ix);
	}

	private void doConn(int ix) {
		int i = (ix >> 8);
		String ipa = String.format("%d.%d.%d.%d",
			sock_regs[ix + Sn_DIPR] & 0xff,
			sock_regs[ix + Sn_DIPR+1] & 0xff,
			sock_regs[ix + Sn_DIPR+2] & 0xff,
			sock_regs[ix + Sn_DIPR+3] & 0xff);
		int pt = sockGet16(ix + Sn_DPORT);
		try {
			InetAddress ia = InetAddress.getByName(ipa);
			socks[i] = new Socket(ia, pt);
			socki[i] = socks[i].getInputStream();
			socko[i] = socks[i].getOutputStream();
			sock_regs[ix + Sn_SR] = SOCK_ESTABLISHED;
			sock_regs[ix + Sn_IR] |= Sn_IR_CON;
		} catch (Exception ee) {
ee.printStackTrace();
			doDisc(ix);
			// TODO: what is the right error state?
			sock_regs[ix + Sn_SR] = SOCK_CLOSE_WAIT;
		}
		chkIntr(ix);
	}

	private int sockGet16(int ix) {
		return ((sock_regs[ix] & 0xff) << 8) |
			(sock_regs[ix + 1] & 0xff);
	}

	private void sockPut16(int ix, int val) {
		sock_regs[ix] = (byte)(val >> 8);
		sock_regs[ix + 1] = (byte)val;
	}

	private void doDisc(int ix) {
		int i = (ix >> 8);
		if (socks[i] != null) try {
			socks[i].close();
		} catch (Exception ee) {
		}
		socks[i] = null;
		socki[i] = null;
		socko[i] = null;
		sock_regs[ix + Sn_SR] = SOCK_CLOSED;
		sock_regs[ix + Sn_IR] |= Sn_IR_DISCON;
		chkIntr(ix);
	}

	private void doClose(int ix) {
		sock_regs[ix + Sn_SR] = SOCK_CLOSED;
	}

	private void doSend(int ix) {
		startLED();
		int i = (ix >> 8);
		int rr = sockGet16(ix + Sn_TX_RD);
		int wr = sockGet16(ix + Sn_TX_WR);
		if (rr == wr) {
			return; // or... full?
		}
		int a = txbase[i] + rr;
		int n = (wr - rr) & txmask[i];
		int f = (txmask[i] + 1 - n);
		sockPut16(ix + Sn_TX_FSR, f);
		try {
			if (wr < rr) {
				// must split send and wrap
				int n1 = (txmask[i] + 1) - rr;
				socko[i].write(mem, a, n1);
				sockPut16(ix + Sn_TX_RD, 0);
				a = txbase[i];
				n -= n1;
				f += n1;
			}
			socko[i].write(mem, a, n);
			socko[i].flush();
			sockPut16(ix + Sn_TX_RD, wr);
			sock_regs[ix + Sn_IR] |= Sn_IR_SENDOK;
			f = txmask[i] + 1;
		} catch (Exception ee) {
ee.printStackTrace();
			doDisc(ix);
			sock_regs[ix + Sn_IR] |= Sn_IR_TIMEOUT;
		}
		sockPut16(ix + Sn_TX_FSR, f);
		chkIntr(ix);
	}

	// This does not actually receive, it only acknowledges
	// that the host has taken the receive data and we can update pointers.
	// Host has updated Sn_RX_RD to indicate how much data they took.
	private void doRecv(int ix) {
		int i = (ix >> 8);
		int rr = sockGet16(ix + Sn_RX_RD);
		int wr = sockGet16(ix + Sn_RX_WR);
		int len = (wr - rr) & rxmask[i];
		sockPut16(ix + Sn_RX_RSR, len);
		sockPut16(ix + Sn_RX_RD, wr);
//System.err.format("! RSR=%04x RD=%04x WR=%04x\n",
//sockGet16(ix + Sn_RX_RSR),
//sockGet16(ix + Sn_RX_RD),
//sockGet16(ix + Sn_RX_WR));
		// User *must* clear this bit!
		if (len != 0) {
			sock_regs[ix + Sn_IR] |= Sn_IR_RECV;
			chkIntr(ix);
		}
	}

	private void doSock(int ix) {
		int cmd = sock_regs[ix + Sn_CR] & 0xff;
		switch (cmd) {
		case OPEN:
			doOpen(ix);
			break;
		case LISTEN:
			// TODO: implement server mode
			break;
		case CONNECT:
			doConn(ix);
			break;
		case DISCON:
			doDisc(ix);
			break;
		case CLOSE:
			doClose(ix);
			break;
		case SEND:
			doSend(ix);
			break;
		case SEND_MAC:
			// TODO: implement?
			break;
		case SEND_KEEP:
			// TODO: implement?
			break;
		case RECV:
			doRecv(ix);
			break;
		}
		// This indicates "command complete"
		sock_regs[ix + Sn_CR] = IDLE;
	}

	private void dumpMsg(String tag, byte[] msg, int start, int len) {
		for (int x = 0; x < len; ++x) {
			System.err.format("%s%02x", tag, msg[start + x] & 0xff);
			tag = " ";
		}
	}

	// Non-blocking!
	private void recv(int i) {
		int ix = (i << 8);
		if (socks[i] == null) {
			return;
		}
		try {
			if (socki[i].available() <= 0) {
				return;
			}
		} catch (Exception ee) {
ee.printStackTrace();
			doDisc(ix);
			return;
		}
		startLED();
		int wr = sockGet16(ix + Sn_RX_WR);
		wr &= rxmask[i];
		int n = (rxmask[i] + 1) - wr;
		int a = rxbase[i] + wr;
		int l = -1;
		try {
			l = socki[i].read(mem, a, n);
			if (l <= 0) {
				return;
			}
//dumpMsg("<", mem, a, l);
			if (l == n && socki[i].available() > 0) {
				wr = 0;
				a = rxbase[i] + wr;
				l = (rxmask[i] + 1) - n;
				l = socki[i].read(mem, a, l);
				if (l > 0) {
//dumpMsg(":", mem, a, l);
					wr += l;
					l += n;
				} else {
					l = n;
				}
			} else {
				wr += l;
			}
//System.err.format("\n");
			wr &= rxmask[i];
			sockPut16(ix + Sn_RX_WR, wr);
			// TODO: may need synchronize() {...}
			// overflow possible?
			sockPut16(ix + Sn_RX_RSR,
				sockGet16(ix + Sn_RX_RSR) + l);
			sock_regs[ix + Sn_IR] |= Sn_IR_RECV;
			chkIntr(ix);
//System.err.format("RSR=%04x RD=%04x WR=%04x l=%02x\n",
//sockGet16(ix + Sn_RX_RSR),
//sockGet16(ix + Sn_RX_RD),
//sockGet16(ix + Sn_RX_WR), l);
		} catch (Exception ee) {
ee.printStackTrace();
			doDisc(ix);
		}
	}

	class Receiver implements Runnable {
		public void run() {
			while (true) {
				for (int x = 0; x < nsocks; ++x) {
					recv(x);
				}
				try {
					Thread.sleep(10);
				} catch (Exception ee) {}
			}
		}
	}

	public void run() {
		while (true) {
			int ix = -1;
			try {
				ix = cmd.take();
			} catch (Exception ee) { }
			if (ix < 0) {
				continue;
			}
			doSock(ix);
		}
	}

	public String getDeviceName() { return name; }
	public String dumpDebug() {
		String ret = String.format("Port %02x IRQ %d\n", base, irq);
		ret += String.format("MR = %02x  ADR = %04x\n",
			com_regs[MR] & 0xff, adr);
		ret += String.format("IR = %02x  IMR = %02x\n",
			com_regs[IR] & 0xff,
			com_regs[IMR] & 0xff);
		ret += String.format("SIPR = %d.%d.%d.%d\n",
			com_regs[SIPR] & 0xff,
			com_regs[SIPR+1] & 0xff,
			com_regs[SIPR+2] & 0xff,
			com_regs[SIPR+3] & 0xff);
		ret += String.format("SUBR = %d.%d.%d.%d\n",
			com_regs[SUBR] & 0xff,
			com_regs[SUBR+1] & 0xff,
			com_regs[SUBR+2] & 0xff,
			com_regs[SUBR+3] & 0xff);
		ret += String.format("GAR  = %d.%d.%d.%d\n",
			com_regs[GAR] & 0xff,
			com_regs[GAR+1] & 0xff,
			com_regs[GAR+2] & 0xff,
			com_regs[GAR+3] & 0xff);
		ret += String.format("SHAR = %02x:%02x:%02x:%02x:%02x:%02x\n",
			com_regs[SHAR] & 0xff,
			com_regs[SHAR+1] & 0xff,
			com_regs[SHAR+2] & 0xff,
			com_regs[SHAR+3] & 0xff,
			com_regs[SHAR+4] & 0xff,
			com_regs[SHAR+5] & 0xff);
		ret += String.format("PMAGIC = %02x\n", com_regs[PMAGIC] & 0xff);
		ret += String.format("RMSR = %dK %dK %dK %dK\n",
			1 << (com_regs[RMSR] & 3),
			1 << ((com_regs[RMSR] >> 2) & 3),
			1 << ((com_regs[RMSR] >> 4) & 3),
			1 << ((com_regs[RMSR] >> 6) & 3));
		ret += String.format("TMSR = %dK %dK %dK %dK\n",
			1 << (com_regs[TMSR] & 3),
			1 << ((com_regs[TMSR] >> 2) & 3),
			1 << ((com_regs[TMSR] >> 4) & 3),
			1 << ((com_regs[TMSR] >> 6) & 3));
		for (int x = 0; x < nsocks; ++x) {
			int ix = (x << 8);
			ret += String.format("S%d_MR = %02x CR = %02x IR = %02x SR = %02x\n",
				x,
				sock_regs[ix + Sn_MR] & 0xff,
				sock_regs[ix + Sn_CR] & 0xff,
				sock_regs[ix + Sn_IR] & 0xff,
				sock_regs[ix + Sn_SR] & 0xff);
			ret += String.format("   DIPR = %d.%d.%d.%d DPORT = %d (%04x)\n",
				sock_regs[ix + Sn_DIPR] & 0xff,
				sock_regs[ix + Sn_DIPR+1] & 0xff,
				sock_regs[ix + Sn_DIPR+2] & 0xff,
				sock_regs[ix + Sn_DIPR+3] & 0xff,
				sockGet16(ix + Sn_DPORT),
				sockGet16(ix + Sn_PORT));
			ret += String.format("   TX_FSR = %04x RD = %04x WR = %04x\n",
				sockGet16(ix + Sn_TX_FSR),
				sockGet16(ix + Sn_TX_RD),
				sockGet16(ix + Sn_TX_WR));
			ret += String.format("   RX_RSR = %04x RD = %04x (WR = %04x)\n",
				sockGet16(ix + Sn_RX_RSR),
				sockGet16(ix + Sn_RX_RD),
				sockGet16(ix + Sn_RX_WR));
		}
		return ret;
	}
}
