// Copyright (c) 2020 Douglas Miller <durgadas311@gmail.com>

import java.io.FileInputStream;
import java.util.Properties;
import java.io.*;

public class Z180Memory extends H89Roms implements Memory, GppListener {
	GeneralPurposePort gpp;
	int gpioBit;
	int gpioORG0;
	// One single bank of 512K, plus ROMs...
	// ROM is fully mapped into 0xf8000-0xfffff, !ORG0-mapped traditionally.
	private byte[] mem;
	private boolean rom;

	// Only supports 27/28C256...
	public Z180Memory(Properties props, GeneralPurposePort gpio, Interruptor intr) {
		super(props, gpio, intr);
		gpp = gpio;
		// rely entirely on notifications for 'rom' value.
		gpioORG0 = gpp.getOrg0Bit();
		gpioBit = romBit | gpioORG0;
		gpp.addGppListener(this);
		mem = new byte[512*1024];
	}

	public int read(boolean rom, int bank, int address) {
		address &= 0xfffff;
		if (address >= 0xf8000) {
			return mon.read(address);
		}
		if (rom && (address < 0x01000 || 
				(address >= 0x01800 && address < 0x02000))) {
			return mon.read(address);
		}
		if (address >= 0x80000) return 0; // or 0xff?
		return mem[address] & 0xff;
	}
	public int read(int address) {
		return read(rom, 0, address);
	}
	public void write(int address, int value) {
		address &= 0xfffff;
		if (address >= 0xf8000) mon.write(address, value);
		if (address >= 0x80000) return;
		mem[address] = (byte)value;
	}

	public void reset() {}

	public void dumpCore(String file) {
		try {
			OutputStream core = new FileOutputStream(file);
			core.write(mem);
			core.close();
		} catch (Exception ee) {
			H89Operator.error(null, "Core Dump", ee.getMessage());
		}
	}

	public int interestedBits() { return gpioBit; }
	public void gppNewValue(int gpio) {
		rom = ((gpio & gpioORG0) == 0);
		if (romBit != 0) romGpp(gpio);
	}
}
